package de.uni_mannheim.app.context;

import java.util.Date;

import de.uni_mannheim.app.business.Job;
import de.uni_mannheim.app.business.Status;
import de.uni_mannheim.app.controllers.AbstractController;
import de.uni_mannheim.app.controllers.InCarController;
import de.uni_mannheim.aware.api.BooleanCondition;
import de.uni_mannheim.aware.api.BooleanConditionListener;
import de.uni_mannheim.aware.api.Condition;
import de.uni_mannheim.aware.api.Rule;
import de.uni_mannheim.aware.api.RuleListener;
import de.uni_mannheim.aware.api.SensorFactory;
import de.uni_mannheim.aware.api.bluetooth.BluetoothDevice;
import de.uni_mannheim.aware.api.bluetooth.BluetoothDeviceAvailableCondition;
import de.uni_mannheim.aware.api.custom.CustomDataEqualsCondition;
import de.uni_mannheim.aware.api.custom.CustomDataNotEqualsCondition;
import de.uni_mannheim.aware.api.custom.CustomSensor;
import de.uni_mannheim.aware.api.location.LocationCondition;
import de.uni_mannheim.aware.api.location.LocationData;
import de.uni_mannheim.aware.api.location.LocationNearCondition;
import de.uni_mannheim.aware.api.location.LocationSensor;
import de.uni_mannheim.aware.api.location.SpeedCondition;
import de.uni_mannheim.aware.api.time.TimeCondition;
import de.uni_mannheim.aware.api.time.TimeData;
import de.uni_mannheim.aware.impl.CommonCustomSensor;
import de.uni_mannheim.aware.impl.SensorNotAvailableException;
import de.uni_mannheim.aware.impl.UnsupportedPlatformException;

public class RuleContext implements Context {
	private static RuleContext instance;
	
	// References
	private ControllerContext controllerContext;
	private WorkerContext workerContext;
	
	// Sensors
	private CustomSensor businessSensor;
	private LocationSensor locationSensor;
	
	// Rules and Conditions
	private Rule inCarDetectionRule;
	private Rule inCarDetectionDrivingRule;
	private Rule atCustomerRule;
	private Rule somewhereRule;
	private Rule lateArrivalRule;
	
	private BooleanCondition atCustomerCondition;
	private BooleanCondition bluetoothCarCondition;	
	private BooleanCondition statusAssignedCondition;
	private BooleanCondition notAtCustomerCondition;
	private BooleanCondition notInCarCondition;
	private BooleanCondition notMovingCondition;
	private BooleanCondition movingCondition;
	private BooleanCondition lateArrivalCondition;
	
	
	private RuleContext() {
		// Fressen!
	}
	
	public synchronized static RuleContext getInstance() {
		if (instance == null)
			instance = new RuleContext();
		
		return instance;
	}	
	
	public void create() {
		try {
			// Sensors
			businessSensor = SensorFactory.getImpl().getCustomSensor();		
			locationSensor = SensorFactory.getImpl().getLocationSensor();
			
			// Static/Fixed Conditions
			atCustomerCondition = new LocationCondition() {				
				protected boolean matches(LocationData currentLocation) {
					if (workerContext.getCurrentJob() == null) {
						System.out.println("current job is NULLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL");return false; }
					
					LocationData jobLocation = 
						new LocationData(workerContext.getCurrentJob().getLatitide(), workerContext.getCurrentJob().getLongitude());
										
					System.out.println(" dist = "+currentLocation.distance(jobLocation)+" --- LOOOOOOOOOCCC "+jobLocation+" ----- my loc: "+currentLocation);
					return currentLocation.distance(jobLocation) <= 100;
				}
			};			
			
			bluetoothCarCondition = new BluetoothDeviceAvailableCondition(new BluetoothDevice("002556D0148C") /* siehe emu-bt sensor */);
			statusAssignedCondition = new CustomDataEqualsCondition("status", Status.ASSIGNED);
			notAtCustomerCondition = new CustomDataNotEqualsCondition("atCustomer", new Boolean(true));
			notInCarCondition = new CustomDataNotEqualsCondition("inCar", new Boolean(true));
			notMovingCondition = new SpeedCondition(30, SpeedCondition.LESS_THAN);
			movingCondition = new SpeedCondition(30, SpeedCondition.GREATER_THAN);
			
			lateArrivalCondition = new TimeCondition() {		
				private int BUFFER_SECONDS = 0;
				
				protected boolean matches(TimeData timeData) {
					if (workerContext.getCurrentJob() == null || workerContext.getEstimatedArrivalTime() == null) {
						System.out.println("[lateArrivalCondition] No job or estimated time known."); return false; }					
					
					long estArrivalTime = workerContext.getEstimatedArrivalTime().getTime();
					long jobStartTime = workerContext.getCurrentJob().getEstimatedStartTime().getTime()
						+ BUFFER_SECONDS*1000;						
					
					boolean lateArrival = estArrivalTime > jobStartTime;
					
					System.out.println("[lateArrivalCondition] Late Arrival? --> "+lateArrival);
					System.out.println("[lateArrivalCondition] jobStart = "+new Date(jobStartTime)+" VS. estArrival = "+new Date(estArrivalTime));
					return lateArrival;							
				}
			};
						
			// In car (not driving)
			inCarDetectionRule = new Rule();
			
			inCarDetectionRule.addCondition(bluetoothCarCondition);
			inCarDetectionRule.addCondition(statusAssignedCondition);
			inCarDetectionRule.addCondition(notAtCustomerCondition);
			inCarDetectionRule.addCondition(notMovingCondition);			
			
			// In car (driving)
			inCarDetectionDrivingRule = new Rule();
			
			//inCarDetectionDrivingRule.addCondition(bluetoothCarCondition);
			inCarDetectionDrivingRule.addCondition(statusAssignedCondition);
			inCarDetectionDrivingRule.addCondition(notAtCustomerCondition);
			inCarDetectionDrivingRule.addCondition(movingCondition);
			
			// Job assigned (not in car, not at customer)
			somewhereRule = new Rule();
			
			somewhereRule.addCondition(statusAssignedCondition);
			somewhereRule.addCondition(notInCarCondition);
			somewhereRule.addCondition(notAtCustomerCondition);
			somewhereRule.addCondition(notMovingCondition);
			
			// Late arrival
			lateArrivalRule = new Rule();
			lateArrivalRule.addCondition(lateArrivalCondition);
			
			// Dynamic rules (empty; updated when data available)
			atCustomerRule = new Rule();
									
			atCustomerCondition.addConditionListener(new BooleanConditionListener() {			
				public void conditionChanged(Condition condition, boolean matches) {
					System.out.println("[atCustomerCondition] Updating business sensor: atCustomer = "+matches);
					businessSensor.put("atCustomer", new Boolean(matches)); 
				}
			});		
			
			atCustomerRule.addCondition(atCustomerCondition);
			atCustomerRule.addCondition(statusAssignedCondition);			
		}
		catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void inject() {
		controllerContext = ControllerContext.getInstance();
		workerContext = WorkerContext.getInstance();
		
		// Sensors
		businessSensor.addListener(controllerContext.getBusinessSensorController());
		locationSensor.addListener(controllerContext.getLocationSensorController());
		
		// Listeners
		inCarDetectionRule.addRuleListener(controllerContext.getInCarController());		
		inCarDetectionDrivingRule.addRuleListener(controllerContext.getInCarDrivingController());
		somewhereRule.addRuleListener(controllerContext.getJobAssignedController());

		// Listeners (also for dynamic rules!)
		atCustomerRule.addRuleListener(controllerContext.getAtCustomerController());
		lateArrivalRule.addRuleListener(controllerContext.getLateArrivalController());
		
		// For updating the business sensor 'inCar' + 'atCustomer'		
		bluetoothCarCondition.addConditionListener(new BooleanConditionListener() {			
			public void conditionChanged(Condition condition, boolean matches) {
				System.out.println("[bluetoothCarConditionListener] "+condition);
				businessSensor.put("inCar", new Boolean(matches));				
			}
		});
		
		movingCondition.addConditionListener(new BooleanConditionListener() {			
			public void conditionChanged(Condition condition, boolean matches) {
				System.out.println("[movingConditionListener] "+condition);
			}
		});
		
		notAtCustomerCondition.addConditionListener(new BooleanConditionListener() {
			
			public void conditionChanged(Condition condition, boolean matches) {
				System.out.println("[notAtCustomerConditionListener] "+condition);				
			}
		});
	}
	
	public void activate() {
		// Activate static/fixed rules
		inCarDetectionRule.setActive(true);
		inCarDetectionDrivingRule.setActive(true);
		somewhereRule.setActive(true);
		atCustomerRule.setActive(true);		
		
		// Do not activate rules that cannot be known yet
		lateArrivalRule.setActive(false); // activated by GetWorkerContextController										
		

		// TODO ugggggggglyyyyyyyyy
		// TODO ugggggggglyyyyyyyyy
		// TODO ugggggggglyyyyyyyyy
		
		/***
		 * 
		 * 
		 *    UGGGLLLY WORKAROUND
		 *
		 *
		 *   normally the sensor should update the conditions when activating
		 *   a rule.
		 *   */
		movingCondition.sensorDataChanged(locationSensor.getLocation());
		notMovingCondition.sensorDataChanged(locationSensor.getLocation());
		atCustomerCondition.sensorDataChanged(locationSensor.getLocation());
		
		businessSensor.put("atCustomer", new Boolean(atCustomerCondition.isConditionMet()));				
		
	}
	
	
	public synchronized void updateAtCustomerRule(Job job) {
		// siehe unten!
	}
	
	public synchronized void updateAtCustomerRule_OOOLDD(Job job) {
		System.out.println("[RuleContext] Updating dynamic AT CUSTOMER rule for job "+job);
		
		// Deactivate old rule				
		atCustomerRule.setActive(false);
		atCustomerRule.removeConditions();
				
		// Update customer rule
		LocationData jobLocation = new LocationData(job.getLatitide(), job.getLongitude());

		try {
			BooleanCondition atCustomerCondition = new LocationCondition() {				
				protected boolean matches(LocationData currentLocation) {
					LocationData custLoc = (LocationData) workerContext.getLocation().clone();
					return currentLocation.distance(custLoc) <= 30;
				}
			};
			
			atCustomerCondition.addConditionListener(new BooleanConditionListener() {			
				public void conditionChanged(Condition condition, boolean matches) {
					System.out.println("[atCustomerCondition] Updating business sensor: atCustomer = "+matches);
					businessSensor.put("atCustomer", new Boolean(matches)); 
				}
			});		
			
			atCustomerRule.addCondition(atCustomerCondition);
			atCustomerRule.addCondition(statusAssignedCondition);
			
			// Set business sensor with current value
			System.out.println("UGGGLLLY WORKAROUND" +
					" -- setting atCustomerCondition manually "+
					" -- normally the sensor should do that when calling " +
					"setActive() t");
			
			// TODO ugggggggglyyyyyyyyy
			atCustomerCondition.sensorDataChanged(locationSensor.getLocation());

			// TODO ugggggggglyyyyyyyyy
			// TODO ugggggggglyyyyyyyyy
			// TODO ugggggggglyyyyyyyyy
			
			/***
			 * 
			 * 
			 *    UGGGLLLY WORKAROUND
			 *
			 *
			 *   normally the sensor should update the conditions when activating
			 *   a rule.
			 *   
			 * 
			 * 
			 * 
			 */			

			// Set business sensor with current value
			businessSensor.put("atCustomer", new Boolean(atCustomerCondition.isConditionMet()));				
	
		} 
		catch (Exception e) {
			System.out.println(e);
			// TODO ...
			e.printStackTrace();
		}		

		System.out.println("[RuleContext] AT CUSTOMER rule activated: "+atCustomerRule);
		atCustomerRule.setActive(true);
		
	
	}

	public CustomSensor getBusinessSensor() {
		return businessSensor;
	}

	public void setBusinessSensor(CommonCustomSensor businessSensor) {
		this.businessSensor = businessSensor;
	}
	
	public Rule getInCarDetectionRule() {
		return inCarDetectionRule;
	}

	public void setInCarDetectionRule(Rule inCarDetectionRule) {
		this.inCarDetectionRule = inCarDetectionRule;
	}

	public Rule getInCarDetectionDrivingRule() {
		return inCarDetectionDrivingRule;
	}

	public void setInCarDetectionDrivingRule(Rule inCarDetectionDrivingRule) {
		this.inCarDetectionDrivingRule = inCarDetectionDrivingRule;
	}

	public Rule getAtCustomerRule() {
		return atCustomerRule;
	}

	public void setAtCustomerRule(Rule atCustomerRule) {
		this.atCustomerRule = atCustomerRule;
	}

	public Condition getBluetoothCarCondition() {
		return bluetoothCarCondition;
	}

	public void setBluetoothCarCondition(BooleanCondition bluetoothCarCondition) {
		this.bluetoothCarCondition = bluetoothCarCondition;
	}

	public Condition getStatusAssignedCondition() {
		return statusAssignedCondition;
	}

	public void setStatusAssignedCondition(BooleanCondition statusAssignedCondition) {
		this.statusAssignedCondition = statusAssignedCondition;
	}

	public Condition getNotMovingCondition() {
		return notMovingCondition;
	}

	public void setNotMovingCondition(BooleanCondition notMovingCondition) {
		this.notMovingCondition = notMovingCondition;
	}

	public Condition getMovingCondition() {
		return movingCondition;
	}

	public void setMovingCondition(BooleanCondition movingCondition) {
		this.movingCondition = movingCondition;
	}

	public Rule getSomewhereRule() {
		return somewhereRule;
	}

	public void setSomewhereRule(Rule somewhereRule) {
		this.somewhereRule = somewhereRule;
	}

	public LocationSensor getLocationSensor() {
		return locationSensor;
	}

	public void setLocationSensor(LocationSensor locationSensor) {
		this.locationSensor = locationSensor;
	}

	public Rule getLateArrivalRule() {
		return lateArrivalRule;
	}

	public void setLateArrivalRule(Rule lateArrivalRule) {
		this.lateArrivalRule = lateArrivalRule;
	}

	
	
}
