package de.uni_mannheim.aware.api;

import java.util.Enumeration;
import java.util.Vector;

import de.uni_mannheim.aware.util.StringUtil;
import de.uni_mannheim.aware.util.VectorUtil;

public class Rule implements BooleanConditionListener {
	private Vector listeners;
	private Vector conditions;
	private boolean ruleSatisfied;
	private boolean active;
	
	public Rule() {
		listeners = new Vector(); 
		conditions = new Vector(); 
		ruleSatisfied = false;
		active = false;
	}
	
	public Rule(Condition[] conditions) {
		this();
		
		for (int i=0; i<conditions.length; i++)
			addCondition(conditions[i]);
	}
	
	public void addRuleListener(RuleListener listener) {
		synchronized (listeners) {
			if (listeners.contains(listener))
				return;
			
			listeners.addElement(listener);			
		}
	}
	
	public boolean removeRuleListener(RuleListener listener) {
		synchronized (listeners) {
			return listeners.removeElement(listener);			
		}
	}
	
	public synchronized void addCondition(Condition condition) {
		if (conditions.contains(condition))
			return;
		
		conditions.addElement(condition);							
	}
	
	public synchronized boolean removeCondition(Condition condition) {
		return conditions.removeElement(condition);			
	}
	
	public synchronized void removeConditions() {
		while (conditions.size() > 0)
			conditions.removeElementAt(0);
	}
	
	public synchronized boolean isRuleSatisfied() {
		return ruleSatisfied;			
	}
	
	public synchronized void setRuleSatisfied(boolean ruleSatisfied) {
		this.ruleSatisfied = ruleSatisfied;
	}
	
	/**
	 * Activates/deactivates the rule by activating/deactivating all attached
	 * conditions, and registering/deregistering itself to them as a listener.
	 * 
	 * <p>If <code>active</code> is set to TRUE, the rule may issue {@link RuleListener#ruleChanged}
	 * events once all conditions are met. 
	 * 
	 * <p>If <code>active</code> is set to FALSE, all corresponding conditions will be detached from the
	 * sensors. Unused sensors will be stopped to save energy.
	 * 
	 * @param active TRUE to enable the rule, FALSE to disable it
	 */
	public synchronized void setActive(boolean active) {
		if (this.active == active)
			return;
		
		// Value changed 
		this.active = active;
		
		for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
			BooleanCondition c = (BooleanCondition) e.nextElement();

			// Add / remove listeners
			c.setActive(active);
			
			if (active)
				c.addConditionListener(this);
			
			else
				c.removeConditionListener(this);
			
			
			
			
			
			// TODO update condition data
			
			/*
			 * If the condition is activated and the sensor doesn't deliver new
			 * data frequently, the condition might hold the wrong state
			 * 
			 * e.g.
			 * 1. location = 49 / 8 --> now loc provider becomes unavailable (no updates!)
			 * 2. now activate rule/condition
			 * 3. condition can't know the state
			 * 
			 * 
			 */
			
			// TODO
			
			
			
			
			
		}		
	}
	
	public synchronized boolean isActive() {
		return active;
	}
	
	/* @Override */
	public synchronized void conditionChanged(Condition condition, boolean matches) {
		// Check all conditions if met
		int matchCount = 0;
		
		for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
			BooleanCondition c = (BooleanCondition) e.nextElement();
			
			if (c.isConditionMet())
				matchCount++;
		}
		
		// If status changed, issue event
		boolean ruleMatches = matchCount == conditions.size();
		System.out.println("matchCount="+matchCount+", matches="+ruleMatches);

		if (isRuleSatisfied() != ruleMatches) {
			setRuleSatisfied(ruleMatches);
			fireRuleChangedEvent();
		}
	}

	private void fireRuleChangedEvent() {
		//System.out.println("fireevent matches="+matches);
		synchronized (listeners) {
			for (Enumeration e = listeners.elements(); e.hasMoreElements(); )				
				((RuleListener) e.nextElement()).ruleChanged(this, ruleSatisfied);
		}	
		
	}
	
	/* @Override */
	public String toString() {
		String conditions = StringUtil.join(",", VectorUtil.toArray(this.conditions));
		return "{matches:"+ruleSatisfied+",conditions:["+conditions+"]}";
	}
}
