package de.uni_mannheim.aware.api;

import java.util.Enumeration;
import java.util.Vector;


/**
 * A RuleSet consists of several rules, that can be added using the 
 * addRule() method. The RuleSet only becomes active after a call
 * of setActive(true). This method will enable all sensors used by the underlying conditions
 * effectively starting the condition sensing.
 * <br/>
 * The RuleSet becomes true when at least one of it's rules is true. Only when
 * all the contained rules are false, the RuleSet is also false. Logically, the
 * rules are put together by an OR. 
 * @author Lars Bakker
 */
public class RuleSet implements RuleListener{
	/**
	 * A Vector of Rules that make up the RuleSet
	 */
	private Vector rules = new Vector();
	/**
	 * A vector a classes that implement the RuleSetListener interface.
	 * Those classes will get notified when the state of the ruleset changes.
	 */
	private Vector listeners = new Vector();
	/**
	 * Keeps track of the current state of the RuleSet. Defaults to false.
	 */
	private boolean ruleSetSatisfied = false;
	/** 
	 * State variable that that is true, when the RuleSet is active (Sensors work)
	 */
	private boolean active = false;

	/** 
	 * Implementation of the RuleListener interface. Here the RuleSet gets notified of
	 * changes in the containing rules.  
	 */
	public synchronized void ruleChanged(Rule rule, boolean matches) {
		int matchCount = 0;
		
		for (Enumeration e = rules.elements(); e.hasMoreElements(); ) {
			Rule r = (Rule) e.nextElement();
			
			if (r.isRuleSatisfied())
				matchCount++;
		}
		
		// If status changed, issue event
		boolean ruleMatches = matchCount == rules.size();
		//System.out.println("matchCount="+matchCount+", changed="+ruleChanged+", matches="+ruleMatches);

		if (isRuleSetSatisfied() != ruleMatches) {
			setRuleSetSatisfied(ruleMatches);
			fireRuleSetChangedEvent();
		}
		
	}
	
	/**
	 * Adds a rule to this RuleSet. If one of the contained rules become true,
	 * the whole RuleSet will also become true and notify it's listeners.
	 * @param rule the Rule to add to this RuleSet
	 */
	public void addRule(Rule rule){
		synchronized (rules) {
			if (rules.contains(rule))
				return;
			
			rules.addElement(rule);
		}		
	}
	
	/**
	 * Remove a rule that was accociated with this RuleSet previously.
	 * @param rule the Rule to remove from this RuleSet
	 * @return true if the rule was contained in this RuleSet and successfully removed, false otherwise
	 */
	public boolean removeRule(Rule rule){
		synchronized (rules) {
			return rules.removeElement(rule);			
		}
	}
	
	/**
	 * Adds a listener to this RuleSet. The listener will get updated when
	 * the state of the RuleSet changes
	 * @param listener The listener that has to be notified when the RuleSet changes
	 */
	public void addSetRuleListener(RuleSetListener listener) {
		synchronized (listeners) {
			if (listeners.contains(listener))
				return;
			
			listeners.addElement(listener);			
		}
	}
	
	/** 
	 * Removes the listener from the list
	 * @param listener 
	 * @return true is successfully removed
	 */
	public boolean removeSetRuleListener(RuleSetListener listener) {
		synchronized (listeners) {
			return listeners.removeElement(listener);			
		}
	}

	/**
	 * If set to true. this method starts sensing for condition updates
	 * @param active true to start listening, false to stop listening
	 */
	public synchronized void setActive(boolean active) {
		if (this.active == active)
			return;
		
		// Value changed 
		this.active = active;
		
		for (Enumeration e = rules.elements(); e.hasMoreElements(); ) {
			Rule r = (Rule) e.nextElement();

			// Add / remove listeners
			r.setActive(active);
			
			if (active)
				r.addRuleListener(this);
			
			else
				r.removeRuleListener(this);
		}		
	}
	
	/**
	 * State variable
	 * @return true if the RuleSet is active, false otherwise
	 */
	public synchronized boolean isActive() {
		return active;
	}

	/**
	 * Notifiy all listeners, that this RuleSet has changed it's state
	 */
	private void fireRuleSetChangedEvent() {
		synchronized (listeners) {
			for (Enumeration e = listeners.elements(); e.hasMoreElements(); )				
				((RuleSetListener) e.nextElement()).ruleSetChanged(this, ruleSetSatisfied);
		}	
		
	}

	/**
	 * True is one of the contained Rules is true, false if alle the contained rules are also false.
	 * @return true if this RuleSet is satisfied, false otherwise
	 */
	public boolean isRuleSetSatisfied() {
		return ruleSetSatisfied;
	}

	/**
	 * True is one of the contained Rules is true, false if alle the contained rules are also false.
	 * @param ruleSetSatisfied true if this RuleSet is satisfied, false otherwise
	 */
	private void setRuleSetSatisfied(boolean ruleSetSatisfied) {
		this.ruleSetSatisfied = ruleSetSatisfied;
	}
	
}
