package de.uni_mannheim.aware.api;

import de.uni_mannheim.aware.api.battery.BatterySensor;
import de.uni_mannheim.aware.api.bluetooth.BluetoothSensor;
import de.uni_mannheim.aware.api.custom.CustomSensor;
import de.uni_mannheim.aware.api.location.LocationSensor;
import de.uni_mannheim.aware.api.location.gps.GPSSensor;
import de.uni_mannheim.aware.api.movement.MovementSensor;
import de.uni_mannheim.aware.api.noise.NoiseLevelSensor;
import de.uni_mannheim.aware.api.time.TimeSensor;
import de.uni_mannheim.aware.api.wireless.WirelessSensor;
import de.uni_mannheim.aware.impl.CommonBluetoothSensor;
import de.uni_mannheim.aware.impl.CommonCustomSensor;
import de.uni_mannheim.aware.impl.CommonGPSSensor;
import de.uni_mannheim.aware.impl.CommonNoiseLevelSensor;
import de.uni_mannheim.aware.impl.CommonTimeSensor;
import de.uni_mannheim.aware.impl.UnsupportedPlatformException;
import de.uni_mannheim.aware.impl.SensorNotAvailableException;
import de.uni_mannheim.aware.impl.emu.EmulatedSensorFactory;
import de.uni_mannheim.aware.impl.nokia.NokiaSensorFactory;

public abstract class SensorFactory {
	public static final byte PLATFORM_UNKNOWN = 0;
	public static final byte PLATFORM_EMULATOR = 1;
	public static final byte PLATFORM_NOKIA = 2;
	public static final byte PLATFORM_RIM = 3;
	public static final byte PLATFORM_SIEMENS = 4;
	
	private static SensorFactory instance;
	
	protected SensorFactory() {

	}	
	
	public static SensorFactory getImpl() throws UnsupportedPlatformException {
		if (instance != null)
			return instance;
		
		// Determine responsible implementation		
		switch (getPlatform()) {
			case PLATFORM_EMULATOR: 
				instance = new EmulatedSensorFactory();
				return instance;
				
			case PLATFORM_NOKIA: 
				instance = new NokiaSensorFactory(); 
				return instance;
						
			case PLATFORM_RIM:
			case PLATFORM_SIEMENS: 		
			case PLATFORM_UNKNOWN:
			default:
				throw new UnsupportedPlatformException("Platform unknown: "+System.getProperty("microedition.platform"));
				//instance = new EmulatedSensorFactory();
				//return instance;
		}
	}
	
	
	// SENSORS //////////////////////////////////////////////////////////////////////
	
	public TimeSensor getTimeSensor() {
		return CommonTimeSensor.getInstance();
	}
	
	public NoiseLevelSensor getNoiseLevelSensor() throws SensorNotAvailableException {
		return CommonNoiseLevelSensor.getInstance();
	}	
	
	public BluetoothSensor getBluetoothSensor() throws SensorNotAvailableException {
		return CommonBluetoothSensor.getInstance();
	}
	
	public GPSSensor getGPSSensor() throws SensorNotAvailableException {
		return CommonGPSSensor.getInstance();
	}	
	
	public CustomSensor getCustomSensor() {
		return CommonCustomSensor.getInstance();
	}	
		
	
	public abstract LocationSensor getLocationSensor() throws SensorNotAvailableException;
	public abstract MovementSensor getMovementSensor() throws SensorNotAvailableException;
	public abstract WirelessSensor getWirelessSensor() throws SensorNotAvailableException;	
	public abstract BatterySensor getBatterySensor() throws SensorNotAvailableException;

	// PRIVATE HELPER ///////////////////////////////////////////////////////////////
	
	public static byte getPlatform() {
		String platform = System.getProperty("microedition.platform");

		// NOKIA
		if ((platform != null && platform.toLowerCase().indexOf("nokia") != -1)
				|| System.getProperty("com.nokia.mid.batterylevel") != null)
			
			return PLATFORM_NOKIA;
		
		// SIEMENS
		if (platform != null && System.getProperty("com.siemens.IMEI") != null)
			return PLATFORM_SIEMENS;		
		
		// EMULATOR		
		if (platform != null && platform.toLowerCase().indexOf("wtk") != -1) // Wireless Tool Kit
			return PLATFORM_EMULATOR;
		
		if (platform != null && platform.toLowerCase().indexOf("microemulator") != -1) // J2ME Polish
			return PLATFORM_EMULATOR;
				
		return PLATFORM_UNKNOWN;
	}	
}
