package de.uni_mannheim.aware.api.location;

import de.uni_mannheim.aware.api.SensorData;
import de.uni_mannheim.aware.util.MathUtil;

/**
 * Represents the location of the cell phone. The data can be provided
 * by different location providers such as Wifi or GPS.
 * 
 * @author Philipp Heckel
 */
public class LocationData extends SensorData {
	/**
	 * Double values are never equal. In order to determine the 
	 * equality of two objects an accuracy is defined.
	 * 
	 * <p><b>Example</b>:<br />0.012 and 0.015 are equal assuming an accuracy of 0.01,
	 * because <code>abs(0.012-0.015) &lt; 0.01</code>. 
	 * 
	 * @see {@link LocationData#equals(Object)}
	 */
	public static final double ACCURACY = 0.00001; 
	
	
	// NOTE: Update clone()-method if fields are added or altered!
	
	/* y-Axes */
	private double latitude;
	
	/* x-Axes */
	private double longitude;
	private double horizontalAccuracy;
	private double verticalAccuracy;
	private double altitude;
	private double speed;
	private long timestamp;
	
	public LocationData() {
		this(0, 0);
	}
	
	public LocationData(double latitude, double longitude) {
		this(latitude, longitude, Double.MAX_VALUE);
	}	
	
	public LocationData(double latitude, double longitude, double horizontalAccuracy) {
		this(latitude, longitude, horizontalAccuracy, Double.MAX_VALUE);
	}		
	
	public LocationData(double latitude, double longitude, double horizontalAccuracy, double verticalAccuracy) {
		this(latitude, longitude, horizontalAccuracy, Double.MAX_VALUE, 0);
	}		
	
	public LocationData(double latitude, double longitude, double horizontalAccuracy, double verticalAccuracy, double altitude) {
		this(latitude, longitude, horizontalAccuracy, Double.MAX_VALUE, 0, 0);
	}
	
	public LocationData(double latitude, double longitude, double horizontalAccuracy, double verticalAccuracy, double altitude, double speed) {
		this(latitude, longitude, horizontalAccuracy, Double.MAX_VALUE, 0, 0, System.currentTimeMillis());
	}		

	public LocationData(double latitude, double longitude, double horizontalAccuracy, double verticalAccuracy, double altitude, double speed, long timestamp) {
		setLatitude(latitude);
		setLongitude(longitude);
		setHorizontalAccuracy(horizontalAccuracy);
		setVerticalAccuracy(verticalAccuracy);
		setAltitude(altitude);
		setSpeed(speed);
		setTimestamp(timestamp);
	}		
	
	public void setLatitude(double latitude) {
		this.latitude = latitude;
	}
	
	public double getLatitude() {
		return latitude;
	}
	
	public void setLongitude(double longitude) {
		this.longitude = longitude;
	}
	
	public double getLongitude() {
		return longitude;
	}
	
	public void setHorizontalAccuracy(double horizontalAccuracy) {
		this.horizontalAccuracy = horizontalAccuracy;
	}
	
	public double getHorizontalAccuracy() {
		return horizontalAccuracy;
	}
	
	public void setVerticalAccuracy(double verticalAccuracy) {
		this.verticalAccuracy = verticalAccuracy;
	}
	
	public double getVerticalAccuracy() {
		return verticalAccuracy;
	}
	
	public void setAltitude(double altitude) {
		this.altitude = altitude;
	}
	
	public double getAltitude() {
		return altitude;
	}
	
	public void setSpeed(double speed) {
		this.speed = speed;
	}
	
	public double getSpeed() {
		return speed;
	}	
	
	public void setTimestamp(long timestamp) {
		this.timestamp = timestamp;
	}
	
	public long getTimestamp() {
		return timestamp;
	}		
	
	/* @Override */
	public SensorData clone() {
		LocationData clone = new LocationData();
		
		clone.setLatitude(getLatitude());
		clone.setLongitude(getLongitude());
		clone.setHorizontalAccuracy(getHorizontalAccuracy());
		clone.setVerticalAccuracy(getVerticalAccuracy());
		clone.setAltitude(getAltitude());
		clone.setSpeed(getSpeed());
		clone.setTimestamp(getTimestamp());
		
		return clone;
	}
	
	/* @Override */
	public boolean equals(Object obj) {
		if (! (obj instanceof LocationData))
			return false;
		
		LocationData data = (LocationData) obj;
				
		if (Math.abs(data.getLatitude()-getLatitude()) > ACCURACY)
			return false;

		if (Math.abs(data.getLongitude()-getLongitude()) > ACCURACY)
			return false;
		
		if (Math.abs(data.getHorizontalAccuracy()-getHorizontalAccuracy()) > ACCURACY)
			return false;
		
		if (Math.abs(data.getVerticalAccuracy()-getVerticalAccuracy()) > ACCURACY)
			return false;
		
		if (Math.abs(data.getAltitude()-getAltitude()) > ACCURACY)
			return false;	
		
		if (Math.abs(data.getSpeed()-getSpeed()) > ACCURACY)
			return false;			
		
		return true;
	}
	
	/* @Override */
	public String toString() {
		return "{latitude:"+getLatitude()+",longitude:"+getLongitude()+","
			  + "horizontal_accuracy:"+getHorizontalAccuracy()+",vertical_accuracy:"+getVerticalAccuracy()+","
			  + "altitude:"+getAltitude()+",speed:"+getSpeed()+",timestamp:"+getTimestamp()+"}";
	}	
	
	
	/**
	 * This is the earth's mean radius in meters. Using the mean gives the most
	 * accurate results for distances measured with any bearing.
	 * <p>
	 * In truth the earth is not a perfect sphere. The radius of the equator is
	 * 6,378,137 and the polar radius is 6,356,752.3142. The FAI's definition of
	 * 6,371,000 lies between them.
	 * 
	 * <p><b>Taken from:</b><br />
	 * <a href="http://location-source.googlecode.com/svn/trunk/LocationSource/src/java/javax/microedition/location/Coordinates.java">Location Source, an implementation of the Java Location API(JSR-179)</a>
	 */
	private static final double METERS_PER_RADIAN = 6378137;//6371000;
	
	/**
	 * Calculates the geodetic distance between the two points according to the
	 * ellipsoid model of WGS84. Altitude is neglected from calculations.
	 * <p>
	 * The implementation shall calculate this as exactly as it can. However, it
	 * is required that the result is within 0.35% of the correct result.
	 * 
	 * @param to
	 *            - the <code>Coordinates</code> of the destination
	 * @return the distance to the destination in meters
	 * @throws java.lang.NullPointerException
	 *             - if the parameter is <code>null</code>
	 *             
	 * <p><b>Taken from:</b><br />
	 * <a href="http://location-source.googlecode.com/svn/trunk/LocationSource/src/java/javax/microedition/location/Coordinates.java">Location Source, an implementation of the Java Location API(JSR-179)</a>
	 */
	public float distance(LocationData to) {
		if (to == null) {
			throw new IllegalArgumentException(
					"distance does not accept a null parameter.");
		}

		// Convert from degrees to radians.
		double lat1 = Math.toRadians(latitude);
		double lon1 = Math.toRadians(longitude);
		double lat2 = Math.toRadians(to.latitude);
		double lon2 = Math.toRadians(to.longitude);

		// Use the Haversine formula for greater accuracy when measuring
		// short distances.  It is explained in detail here:
		//   http://williams.best.vwh.net/avform.htm
		//   http://www.movable-type.co.uk/scripts/LatLong.html
		// d = 2*asin(sqrt(
		//          (sin((lat1-lat2)/2))^2 +    // d2
		//          cos(lat1)*cos(lat2) *       // d3
		//          (sin((lon1-lon2)/2))^2) )   // d5
		
		double d1 = Math.sin((lat1 - lat2) / 2.0);
		double d2 = d1 * d1;
		double d3 = Math.cos(lat1) * Math.cos(lat2);
		double d4 = Math.sin((lon1 - lon2) / 2.0);
		double d5 = d4 * d4;
		double d6 = d2 + d3 * d5;
		double distanceInRadians = 2.0 * MathUtil.asin(Math.sqrt(d6));

		double distance = METERS_PER_RADIAN * distanceInRadians;
		return (float) distance;

	}

}
