package de.uni_mannheim.aware.impl;

import de.uni_mannheim.aware.api.Config;
import de.uni_mannheim.aware.api.SensorListener;
import de.uni_mannheim.aware.api.bluetooth.BluetoothData;
import de.uni_mannheim.aware.api.bluetooth.BluetoothDevice;
import de.uni_mannheim.aware.api.bluetooth.BluetoothSensor;
import de.uni_mannheim.aware.util.TimedList;
import de.uni_mannheim.aware.util.VectorUtil;
import de.uni_mannheim.aware.util.TimedList.ExpiryListener;

import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.microedition.lcdui.CustomItem;



/**
 * <p><b>Configuration parameters</b>:
 * <ul>
 *   <li><b>bluetooth.sleep</b>: sleep time between the Bluetooth inquiries (in ms).
 *   <li><b>bluetooth.timeout</b>: devices disappear after the given timeout (in ms).
 *   <li><b>bluetooth.timeout-accuracy</b>: defines the accuracy for timeouts (in ms).
 *   If a device times out shortly before another one, it is more efficient
 *   to remove both devices together.
 * </ul>
 * 
 * @see http://www.klings.org/JavaBluetooth/BTBrowser.html
 * @see http://www.nowires.org/Thesis-PDF/AndreK.pdf
 * @author Philipp Heckel
 */
public class CommonBluetoothSensor extends CommonSensor implements BluetoothSensor, ExpiryListener {	
	/**
	 * Sleep time between the Bluetooth inquiries (in ms).
	 * Can be overridden with {@link Config#set(String, String)}
	 * using the key <b>bluetooth.sleep</b>.
	 * 
	 * <p><b>Note</b>: Can be changed during runtime!
	 */
	public static final int DEFAULT_SLEEP_TIME = 1000;
	
	/**
	 * Devices disappear after the given timeout (in ms).
	 * Can be overridden with {@link Config#set(String, String)}
	 * using the key <b>bluetooth.timeout</b>.
	 * 
	 * <p>The timeout must be greater than the 
	 * Bluetooth discovery time. Otherwise, devices might disappear
	 * and reappear again.
	 * 
	 * <p><b>Note</b>: Can be changed during runtime!
	 */
	public static final int DEFAULT_DEVICE_TIMEOUT = 60000;
	
	/**
	 * Defines the accuracy for timeouts: if a device times out shortly
	 * before another one, it is more efficient to remove both devices together.
	 * Can be overridden with {@link Config#set(String, String)}
	 * using the key <b>bluetooth.timeout-accuracy</b>.
	 * 
	 * <p><b>Note</b>: The timeout must be greater than the 
	 * Bluetooth discovery time. Otherwise, devices might disappear
	 * and reappear again.
	 * 
	 * <p><b>Note</b>: <u>Cannot</u> be changed during runtime!
	 */	
	public static final int DEFAULT_DEVICE_TIMEOUT_ACCURACY = 1500;	
	
	private static BluetoothSensor instance;
	
	private DiscoveryAgent discoveryAgent;
	private DiscoveryWorker worker;

	/**
	 * Hold references to the currently known devices, including a specified
	 * expiry date. When the expiry date is reached, an event is thrown.
	 * 
	 * @see {@link CommonSensor#fireSensorDataChangedEvent(de.uni_mannheim.aware.api.SensorData)}
	 */
	private TimedList currentDeviceList;

	private CommonBluetoothSensor() throws SensorNotAvailableException {
		currentDeviceList = new TimedList(Config.getInt("bluetooth.timeout-accuracy", DEFAULT_DEVICE_TIMEOUT_ACCURACY));
		currentDeviceList.addExpiryListener(this);
		
		try {
			discoveryAgent = LocalDevice.getLocalDevice().getDiscoveryAgent();
		}
		catch (BluetoothStateException e) {
			throw new SensorNotAvailableException();
		}
	}
	
	/**
	 * Singleton!
	 * @throws SensorNotAvailableException 
	 */
	public static BluetoothSensor getInstance() throws SensorNotAvailableException {
		if (instance != null)
			return instance;
		
		instance = new CommonBluetoothSensor();
		return instance;
	}		
	
	/* @Override */
	public synchronized void start() {
		if (worker != null)
			return;
		
		worker = new DiscoveryWorker();
		worker.start();
	}
	
	/* @Override */
	public synchronized void stop() {
		if (worker == null)
			return;
		
		worker.stop();
		worker = null;
	}

	/**
	 * Called by the {@link TimedList} when devices vanish, i.e. they haven't
	 * been seen in X seconds.
	 */
	/* @Override */
	public synchronized void objectsExpired(Vector expiredObjects) {
		// Make BluetoothData object (current devices / vanished devices)
		BluetoothData data = new BluetoothData();
		
		data.setDevices(currentDeviceList.getObjects());
		data.setVanished(expiredObjects);

		// Fire!
		fireSensorDataChangedEvent(data);		
	}

	/** 
	 * Called by the {@link DiscoveryWorker} if new devices have been
	 * discovered.
	 * 
	 * @param addedObjects
	 */
	public synchronized void objectsAdded(Vector addedObjects) {
		// Make BluetoothData object (current devices / new devices)
		BluetoothData data = new BluetoothData();
		
		data.setDevices(currentDeviceList.getObjects());
		data.setDiscovered(addedObjects);

		// Fire!
		fireSensorDataChangedEvent(data);	
	}
	
	/* @Override */
	public synchronized BluetoothData getBluetoothData() {
		BluetoothData data = new BluetoothData();		
		data.setDevices(currentDeviceList.getObjects());
		
		return data;
	}

	private class DiscoveryWorker extends Thread implements DiscoveryListener {
		private boolean running;
		private Vector newDeviceList;
		
		public synchronized void stop() {
			running = false;
			discoveryAgent.cancelInquiry(this);

			interrupt();		
		}
		
		public void run() {
			running = true;
			
			// Start discovery loop; stopped by 'interrupt()'-method
			while (running) { 
				try {		
					// Discover synchronously! Blocking!
					// This updates the current devices' time stamps
					// and adds newly discovered ones to the list
					discoverSynchronously();
					
					// Fire notifications for new devices
					synchronized (newDeviceList) {
						if (newDeviceList.size() > 0)
							objectsAdded(newDeviceList);		
					}
					
					// Sleep! Blocking!
					Thread.sleep(Config.getInt("bluetooth.sleep", DEFAULT_SLEEP_TIME));
				}
				catch (BluetoothStateException e) {
					e.printStackTrace();
				}
				catch (InterruptedException e) {
					if (!running)
						return;
				} 
			}			
		}
		
		/**
		 * Discovers available Bluetooth devices and updates {@link CommonBluetoothSensor#currentDeviceList currentDeviceList}
		 * and {@link CommonBluetoothSensor#newDeviceList newDeviceList}.
		 * 
		 * @throws BluetoothStateException
		 * @throws InterruptedException
		 */
		public synchronized void discoverSynchronously() throws BluetoothStateException, InterruptedException {
			newDeviceList = new Vector();
			discoveryAgent.startInquiry(DiscoveryAgent.GIAC, this);
			
			// Wait for async-calls; 'synchronized' is required!
			wait(); 	
		}

		/* @Override */
		public synchronized void deviceDiscovered(RemoteDevice btDevice, DeviceClass cod) {
			BluetoothDevice device = new BluetoothDevice(btDevice.getBluetoothAddress());
			
			// Get friendly name
			try {
				String friendlyName = btDevice.getFriendlyName(false);
				
				if (friendlyName != null)
					device.setFriendlyName(friendlyName);
			} catch (Exception e) {
				// Ignore.
			}
			
			// Add 
			synchronized (currentDeviceList) {
				if (!currentDeviceList.contains(device))
					newDeviceList.addElement(device);				

				// Update existing, or add new entry 
				currentDeviceList.replace(device, Config.getInt("bluetooth.timeout", DEFAULT_DEVICE_TIMEOUT));
			}
		}

		/* @Override */
		public synchronized void inquiryCompleted(int discType) {
			if (discType != DiscoveryListener.INQUIRY_COMPLETED)
				return;
			
			// Wake-up waiting getDevices()-method; 'synchronized' is required!
			notify();	
		}

		/* @Override */
		public void serviceSearchCompleted(int transID, int respCode) {
			// Nothing.
		}

		/* @Override */
		public void servicesDiscovered(int transID, ServiceRecord[] servRecord) {
			// Nothing.
		}
	}



}
