package de.uni_mannheim.aware.impl;

import javax.microedition.location.Criteria;
import javax.microedition.location.Location;
import javax.microedition.location.LocationException;
import javax.microedition.location.LocationListener;
import javax.microedition.location.LocationProvider;

import de.uni_mannheim.aware.api.Config;
import de.uni_mannheim.aware.api.location.LocationData;
import de.uni_mannheim.aware.api.location.gps.GPSSensor;


/**
 * Represents the GPS sensor of the cell phone. The class starts a new
 * thread which permanently captures the current coordinates.
 * 
 * @author Philipp Heckel
 */
public class CommonGPSSensor extends CommonSensor implements GPSSensor, LocationListener {
	/**
	 * Fixed interval for location updates (in seconds). 
	 * <b>-1</b> is used for the default interval of this provider.
	 * <b>0</b> is used to indicate that the application wants to receive only provider.
	 *  
	 * <p>Can be overridden with {@link Config#set(String, String)}
	 * using the key <b>gps.interval</b>.
	 * 
	 * <p><b>Note</b>: The sensor also fires if there is no valid location available.
	 * This value can be changed during runtime!
	 *  
	 * <p><em>"Passing in -1 as the interval selects the default interval which is dependent on the used location method. Passing in 0 as the interval registers the listener to only receive provider status updates and not location updates at all.
	 * [..] Subsequent location updates will happen at the defined interval after the first one. If the specified update interval is smaller than the time it takes to obtain the first result, the listener shall receive location updates with invalid Locations at the defined interval until the first location result is available."</em>
	 * 
	 * @see <a href="http://www-users.cs.umn.edu/~czhou/docs/jsr179/lapi/javax/microedition/location/LocationProvider.html#setLocationListener%28javax.microedition.location.LocationListener,%20int,%20int,%20int%29">LocationProvider.setLocationListener</a>
	 */
	private static final int DEFAULT_INTERVAL = 30;
	
	/**
	 * Timeout for location updates (in seconds). After (interval+timeout) seconds,
	 * an invalid location object will be fired. 
	 * 
	 * <p><em>"Must be greater than 0. 
	 * If the value is -1, the default timeout for this provider is used. 
	 * Also, if the interval is -1 to indicate the default, the value of this 
	 * parameter has no effect and the default timeout for this provider is used.
	 * If the interval is 0, this parameter has no effect."</em>
	 * 
	 * <p>Can be overridden with {@link Config#set(String, String)}
	 * using the key <b>gps.timeout</b>.
	 * 
	 * @see <a href="http://www-users.cs.umn.edu/~czhou/docs/jsr179/lapi/javax/microedition/location/LocationProvider.html#setLocationListener%28javax.microedition.location.LocationListener,%20int,%20int,%20int%29">LocationProvider.setLocationListener</a>
	 */
	private static final int DEFAULT_TIMEOUT = 10;
	
	/**
	 * <em>"The maxAge parameter defines how old the location result is allowed to be provided when the update is made. This allows the implementation to reuse location results if it has a recent location result when the update is due to be delivered. This parameter can only be used to indicate a larger value than the normal time of obtaining a location result by a location method. The normal time of obtaining the location result means the time it takes normally to obtain the result when a request is made. If the application specifies a time value that is less than what can be realized with the used location method, the implementation shall provide as recent location results as are possible with the used location method. For example, if the interval is 60 seconds, the maxAge is 20 seconds and normal time to obtain the result is 10 seconds, the implementation would normally start obtaining the result 50 seconds after the previous update. If there is a location result otherwise available that is more recent than 40 seconds after the previous update, then the maxAge setting to 20 seconds allows to return this result and not start obtaining a new one."</em>
	 * 
	 * <p>Can be overridden with {@link Config#set(String, String)}
	 * using the key <b>gps.maxage</b>.
	 * 
	 * @see <a href="http://www-users.cs.umn.edu/~czhou/docs/jsr179/lapi/javax/microedition/location/LocationProvider.html#setLocationListener%28javax.microedition.location.LocationListener,%20int,%20int,%20int%29">LocationProvider.setLocationListener</a>
	 */
	private static final int DEFAULT_MAXAGE = 30;	
	
	private static CommonGPSSensor instance;
	private static LocationProvider locationProvider;
	
	private LocationData locationData;
	private boolean enabled;
	
	
	public CommonGPSSensor() throws SensorNotAvailableException {
		//throw new SensorNotAvailableException();
		
		// Initialize location provider
		// TODO Make location criteria flexible/configurable
		Criteria cr = new Criteria();		
		
		//cr.setHorizontalAccuracy(10000);
		cr.setSpeedAndCourseRequired(true);
		cr.setCostAllowed(true);
		cr.setAltitudeRequired(true);
		
		try {
			locationProvider = LocationProvider.getInstance(cr);
		}
		catch (LocationException e) {
			throw new SensorNotAvailableException();
		}		
		
		enabled = false;
	}
	
	/**
	 * Singleton!
	 * @throws SensorNotAvailableException 
	 */
	public static GPSSensor getInstance() throws SensorNotAvailableException {
		if (instance != null)
			return instance;
		
		instance = new CommonGPSSensor();
		return instance;
	}		
	
	/* @Override */
	public synchronized void start() {
		if (enabled)
			return;
		
		// Enable!
		locationProvider.setLocationListener(this, 
			Config.getInt("gps.interval", DEFAULT_INTERVAL),
			Config.getInt("gps.timeout", DEFAULT_TIMEOUT),
			Config.getInt("gps.maxage", DEFAULT_MAXAGE));
		
		enabled = true;
	}
	
	/* @Override */
	public synchronized void stop() {
		if (!enabled)
			return;
		
		// Disable!
		locationProvider.setLocationListener(null, -1, -1, -1);
		enabled = false;
	}
	
	public LocationData getLocation() {
		return locationData;
	}

	/* @Override */
	public void locationUpdated(LocationProvider provider, Location location) {
		LocationData newLocationData = new LocationData(
			location.getQualifiedCoordinates().getLatitude(),
			location.getQualifiedCoordinates().getLongitude(),
			location.getQualifiedCoordinates().getHorizontalAccuracy(),
			location.getQualifiedCoordinates().getVerticalAccuracy(),
			location.getQualifiedCoordinates().getAltitude(),
			location.getSpeed()
		);
		
		if (newLocationData.equals(locationData))
			return;
		
		// Changed!
		locationData = newLocationData;
		fireSensorDataChangedEvent(locationData);		
	}
	
	/* @Override */
	public void providerStateChanged(LocationProvider provider, int newState) {
		// TODO Auto-generated method stub
		
	}
	
	
}
