package de.uni_mannheim.aware.impl.emu;

import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;

import de.uni_mannheim.aware.api.SensorData;
import de.uni_mannheim.aware.api.bluetooth.BluetoothSensor;
import de.uni_mannheim.aware.api.location.LocationData;
import de.uni_mannheim.aware.api.location.LocationSensor;
import de.uni_mannheim.aware.impl.CommonSensor;

public class EmulatedLocationPathSensor extends CommonSensor implements LocationSensor {
	private static LocationSensor instance;

	private static int updateInterval = 1000; // milliseconds
	private static double speed = 100; // km/h	
	private static LocationData[] waypoints = new LocationData[] {
		new LocationData(49.481,8.471), // L15
		new LocationData(49.4828,8.47299), // Kaiserring am Cinemax
		new LocationData(49.48637,8.4638), // Nahe Kochschule (Zwischenpunkt)
		new LocationData(49.48765,8.4604), // Kreuzung C5
		new LocationData(49.4873,8.46), // B6
	};
	
	private Timer timer;
	private boolean enabled;
	private int currentWaypointIndex; 
	private double metersPerInterval;
	private LocationData currentLocation;
	private LocationData nextLocation;
	
	private EmulatedLocationPathSensor() {
		this.timer = null;
		this.enabled = false;
		this.currentWaypointIndex = 0;
		this.metersPerInterval = updateInterval * (speed/3.6/1000 /* = meters per millisecond */);
		this.currentLocation = waypoints[currentWaypointIndex];
		this.nextLocation = waypoints[currentWaypointIndex+1];
	}
	
	/**
	 * Singleton!
	 * @return
	 */
	public static LocationSensor getInstance() {
		if (instance != null)
			return instance;
		
		instance = new EmulatedLocationPathSensor();
		return instance;
	}		
	
	/* @Override */
	public synchronized LocationData getLocation() {
		return currentLocation;
	}
	
	/* @Override */
	public synchronized SensorData getSensorData() {
		return getLocation();
	}	
	
	/* @Override */
	public synchronized void start() {
		// Cancel the old one
		if (timer != null)
			return;
		
		// Start timer (= emulated BT discovery!)
		timer = new Timer();
		
		timer.schedule(
			new TimerTask() { public void run() { timerTaskEvent(); } },
			updateInterval, updateInterval);
		
		// Fire one time!
		fireSensorDataChangedEvent(currentLocation);
	}
	
	/* @Override */
	public synchronized void stop() {
		if (timer != null)
			timer.cancel();
	}
	
	/* @Override */
	public boolean isAvailable() {
		return true;
	}
	
	public synchronized void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}
	
	public synchronized boolean getEnabled() {
		return enabled;
	}
	
	private void timerTaskEvent() {
		if (!enabled)
			return;
						
		// No next point: Stop!
		if (currentWaypointIndex == Integer.MAX_VALUE) {
			stop(); return; }
		
		// Get next waypoint
		LocationData middleLocation = calculateNextWaypoint(currentLocation, nextLocation);
		
		// If beyond destination (= nextLocation), 
		if (middleLocation == null) {
			currentWaypointIndex++;

			// This was the last waypoint
			if (currentWaypointIndex == waypoints.length-1) {
				System.out.println("[EmulatedLocationPathSensor] LAST location --> we're at the end of the path.");
				
				currentWaypointIndex = Integer.MAX_VALUE;
				currentLocation = nextLocation;
			}
			
			// It wasn't the last waypoint
			else {				
				currentLocation = waypoints[currentWaypointIndex];
				nextLocation = waypoints[currentWaypointIndex+1];
			}
		}	
		
		// Not beyond destination
		else {
			currentLocation = middleLocation;
		}
		
		currentLocation.setSpeed(speed);
		
		System.out.println("[EmulatedLocationPathSensor] "+speed+" km/h; Location changed: "+currentLocation);
		fireSensorDataChangedEvent(currentLocation);
	}
	
	/**
	 * 
	 * @author Philipp Heckel 
	 * @param a
	 * @param b
	 * @return
	 */
	private LocationData calculateNextWaypoint(LocationData a, LocationData b) {		
		double d = a.distance(b);

		//// Between a and b ////
		
		// Horizontal and vertical direction (required for determining the last point)
		double latdir = (a.getLatitude() > b.getLatitude()) ? -1 : 1;
		double lngdir = (a.getLongitude() > b.getLongitude()) ? -1 : 1;
		
		// Horizontal and vertical distance (in meters); negative distance to move left or down
		double dlat = latdir*a.distance(new LocationData(b.getLatitude(), a.getLongitude()));
		double dlng = lngdir*a.distance(new LocationData(a.getLatitude(), b.getLongitude()));//Math.abs(a.getLongitude()+b.getLongitude());
		
		
		//// Next step (= c) ////
		
		// Horizontal and vertical distance from 'a' to the new point (= c)
		double dnewlat = dlat/d*metersPerInterval;
		double dnewlng = dlng/d*metersPerInterval;
		
		/* Approximate new latitude and longitude
		 * cmp. http://stackoverflow.com/questions/2839533/adding-distance-to-a-gps-coordinate */
		double newlat = a.getLatitude() + (180/Math.PI)*(dnewlat/6378137);
		double newlng = a.getLongitude() + (180/Math.PI)*(dnewlng/6378137)/Math.cos(a.getLatitude());

		LocationData c = new LocationData(newlat, newlng);		
		
		// Check if it is beyond (swap '>' if latdir or lngdir equal -1) 
		//System.out.println(c.getLatitude() +" > "+latdir+" * "+b.getLatitude());
		//System.out.println(c.getLongitude() +" > "+lngdir+" * "+b.getLongitude());
		if (latdir*c.getLatitude() > latdir*b.getLatitude() && lngdir*c.getLongitude() > lngdir*b.getLongitude())
			{ //System.out.println("null");
			return null;}
				
		return c;
	}	
}
