package de.uni_mannheim.aware.impl.nokia;

import java.io.IOException;
import java.util.Date;
import java.util.Timer;

import de.uni_mannheim.aware.api.Config;
import de.uni_mannheim.aware.api.SensorData;
import de.uni_mannheim.aware.api.SensorFactory;
import de.uni_mannheim.aware.api.SensorListener;
import de.uni_mannheim.aware.api.location.LocationData;
import de.uni_mannheim.aware.api.location.LocationSensor;
import de.uni_mannheim.aware.api.location.gps.GPSSensor;
import de.uni_mannheim.aware.api.wireless.WirelessData;
import de.uni_mannheim.aware.api.wireless.WirelessSensor;
import de.uni_mannheim.aware.impl.CommonGPSSensor;
import de.uni_mannheim.aware.impl.CommonSensor;
import de.uni_mannheim.aware.impl.SensorNotAvailableException;
import de.uni_mannheim.aware.impl.UnsupportedPlatformException;
import de.uni_mannheim.aware.impl.emu.EmulatedWirelessSensor;
import de.uni_mannheim.aware.util.GeoUtil;

/**
 * This sensor combines uses wifi data and GPS data to get location data
 * as quickly as possible.
 * 
 * <p>It uses the {@link NokiaWirelessSensor} to retrieve available wireless
 * access points, and queries the Google Geolocation API to get a rough GPS
 * fix. 
 * 
 * <p>To get real GPS data, it uses the {@link CommonGPSSensor}.
 * 
 * <p>Both sensors constantly deliver GPS coordinates, but only the location data
 * with the highest accuracy is used (unless the data is not fresh/up-to-date anymore).
 * 
 * <p><b>Configuration parameters</b>:
 * <ul>
 *   <li>This sensor uses {@link NokiaWirelessSensor} and {@link CommonGPSSensor}.
 *   Refer to their documentation for config parameters.
 *   
 *   <li><b>location.freshness</b>: time in seconds for which a location is considered
 *   up-to-date. If the location is older, another sensor is used (GPS vs. Wifi).
 * </ul>
 * 
 * @see <a href="http://code.google.com/apis/gears/api_geolocation.html">Google Geolocation API</a>
 * @author Philipp Heckel
 */
public class NokiaLocationSensor extends CommonSensor implements LocationSensor {
	private static LocationSensor instance;

	/**
	 * This sensor prefers the most accurate data. But when
	 * the more accurate sensor fails (typically the GPS sensor),
	 * it must use the fall-back sensor (typically wifi).
	 * 
	 * <p>This time specifies the delay before the less accurate data
	 * is used for the sensor.
	 * 
	 * <p>The timeout value van be overridden with {@link Config#set(String, String)}
	 * using the key <b>location.freshness</b> (in seconds).
	 */
	private static final int DEFAULT_FRESHNESS = 10;
	
	private LocationData location;
	private boolean active;

	private WirelessSensor wifiSensor;
	private LocationData wifiLocation;
	private double wifiLocationUpdateTime;
	private SensorListener wifiListener;
	
	private GPSSensor gpsSensor;
	private SensorListener gpsListener;
	private LocationData gpsLocation;
	private double gpsLocationUpdateTime;	

		
	private NokiaLocationSensor() throws SensorNotAvailableException {		
		location = new LocationData();
		active = false;
		
		gpsLocation = new LocationData();
		gpsLocationUpdateTime = 0;
		
		wifiLocation = new LocationData();		
		wifiLocationUpdateTime = 0;
		
		
		// Wifi
		try { 
			wifiSensor = SensorFactory.getImpl().getWirelessSensor();
			//wifiSensor = NokiaWirelessSensor.getInstance();
			wifiListener = new SensorListener() {				
				public void sensorDataChanged(SensorData sensorData) {
					wifiSensorDataChanged((WirelessData) sensorData);
				}
			};
		}
		catch (Exception e) { 
			// Nothing. See below.
		} 
		
		// GPS
		try { 
			gpsSensor = SensorFactory.getImpl().getGPSSensor(); 
			gpsListener = new SensorListener() {				
				public void sensorDataChanged(SensorData sensorData) {
					gpsSensorDataChanged((LocationData) sensorData);
				}
			};
		}
		catch (Exception e) { 
			// Nothing. See below.
		}
		
		// Throw error if none of the above sensors are available.
		if (gpsSensor == null && wifiSensor == null)
			throw new SensorNotAvailableException();
	}
	
	/**
	 * Singleton!
	 * @throws SensorNotAvailableException 
	 */
	public static LocationSensor getInstance() throws SensorNotAvailableException {
		if (instance != null)
			return instance;
		
		instance = new NokiaLocationSensor();
		return instance;
	}

	/* @Override */
	public synchronized void start() {
		if (active)
			return;
		
		active = true;
		
		if (wifiSensor != null)
			wifiSensor.addListener(wifiListener);
		
		if (gpsSensor != null)
			gpsSensor.addListener(gpsListener);
	}
	
	/* @Override */
	public synchronized void stop() {
		if (!active)
			return;
		
		active = false;
		
		if (wifiSensor != null)
			wifiSensor.removeListener(wifiListener);
		
		if (gpsSensor != null)
			gpsSensor.removeListener(gpsListener);
	}

	/**
	 * Called when the {@link GPSSensor} updates its data. If the 
	 * new data does not equal the saved data, an event is fired to 
	 * the listener.
	 * 
	 * @param locationData Updated sensor data
	 */
	private synchronized void gpsSensorDataChanged(LocationData locationData) {
		// Update saved GPS location
		gpsLocation = locationData;		
		gpsLocationUpdateTime = System.currentTimeMillis();
		
		// Figure out which data to use GPS or wifi
		LocationData newLocationData = determineLocation();
		
		// Fire if location data has been updated
		if (!location.equals(newLocationData)) {
			location = newLocationData;
			fireSensorDataChangedEvent(location);	
		}		
	}			
	
	/**
	 * Called when the {@link WirelessSensor} updates its data. 
	 * This method queries Google for a rough fix.
	 * 
	 * @param wirelessData Updated sensor data
	 */
	private void wifiSensorDataChanged(WirelessData wirelessData) {
		// Retrieve GPS coordinates from Google
		try {
			wifiLocation = GeoUtil.getLocation(wirelessData.getDevices());
			wifiLocationUpdateTime = System.currentTimeMillis();
		}
		catch (IOException e) {
			// Ignore, because in this case GPS data is used.
			return;
		}	
		
		// Figure out which data to use GPS or wifi
		LocationData newLocationData = determineLocation();
		
		// Fire if location data has been updated
		if (!location.equals(newLocationData)) {
			location = newLocationData;
			fireSensorDataChangedEvent(location);	
		}				
	}		
	
	/**
	 * Compares the available data (GPS and Wifi) to figure out
	 * which location to use. The implementation always uses the most accurate information
	 * unless it is outdated (timeout).
	 * 
	 * <p>The timeout value van be overridden with {@link Config#set(String, String)}
	 * using the key <b>location.freshness</b> (in seconds).
	 * 
	 * @return data object of the most accurate location / up-to-date location. 
	 */
	private LocationData determineLocation() {
		LocationData newLocationData;
		
		// GPS is more accurate
		if (gpsLocation.getHorizontalAccuracy() < wifiLocation.getHorizontalAccuracy()) {
			// GPS data is up-to-date, use GPS data
			if (System.currentTimeMillis() - gpsLocationUpdateTime < Config.getInt("location.freshness", DEFAULT_FRESHNESS)*1000)
				newLocationData = (LocationData) gpsLocation.clone();
			
			// GPS data is outdated, use wifi data
			else 
				newLocationData = (LocationData) wifiLocation.clone();
		}
		
		// Wifi data is more accurate
		else {
			// Wifi data is up-to-date, use wifi data
			if (System.currentTimeMillis() - wifiLocationUpdateTime < Config.getInt("location.freshness", DEFAULT_FRESHNESS)*1000)
				newLocationData = (LocationData) wifiLocation.clone();
			
			// Wifi data is outdated, use GPS data
			else 
				newLocationData = (LocationData) gpsLocation.clone();			
		}
		
		return newLocationData;
	}
	
	/* @Override */
	public synchronized LocationData getLocation() {
		return location;
	}
		
	/* @Override */
	public synchronized SensorData getSensorData() {
		return getLocation();
	}		

}
