package de.uni_mannheim.aware.util;

public class ByteConverter {
    /**
     * Reads 4 bytes of data representing a signed 32 bit integer value stored as <strong>little endian</strong>. 
     * @param data the byte array where the uint32 has to be extracted
     * @param offset the offset in the byte array of the first byte of the int32 value
     * @return the int32 value
     */
    public static int readIntLittleEndian(byte[] data, int offset){
    	return (((int)data[offset+3] & 0xff)  << 24) | (((int)data[offset+2] & 0xff) << 16) | (((int)data[offset+1]  & 0xff) << 8) | ((int)data[offset] & 0xff);
    }

    /**
     * Reads 4 bytes of data representing a signed 32 bit integer value stored as <strong>big endian</strong>. 
     * @param data the byte array where the uint32 has to be extracted
     * @param offset the offset in the byte array of the first byte of the int32 value
     * @return the int32 value
     */
    public static int readIntBigEndian(byte[] data, int offset){
	    return (((int)data[offset] & 0xff)  << 24) | (((int)data[offset+1] & 0xff) << 16) | (((int)data[offset+2] & 0xff) << 8) | (((int)data[offset+3]) & 0xff);
    }
    
    /**
     * Read 4 bytes of data representing an unsigned 32 bit integer value stored as big. As java only supports
     * signed integers, we have to use the long data type to be able to read the value
     * @param data the byte array where the uint32 has to be extracted
     * @param offset the offset in the byte array of the first byte of the uint32 value
     * @return the uint32 value as long value
     */
    public static long readUnsignedIntBigEndian(byte[] data, int offset){
	    return (((long)data[offset] & 0xffL)  << 24) 
	            | (((long)data[offset+1] & 0xffL) << 16) 
	            | (((long)data[offset+2] & 0xffL) << 8) 
	            | (((long)data[offset+3] & 0xffL) );
    	    	
    }

    /**
     * Read 4 bytes of data representing an unsigned 32 bit integer value stored as little endian. As java only supports
     * signed integers, we have to use the long data type to be able to read the value
     * @param data the byte array where the uint32 has to be extracted
     * @param offset the offset in the byte array of the first byte of the uint32 value
     * @return the uint32 value as long value
     */
    public static long readUnsignedIntLittleEndian(byte[] data, int offset){
	    return (((long)data[offset+3] & 0xffL)  << 24) 
	            | (((long)data[offset+2] & 0xffL) << 16) 
	            | (((long)data[offset+1] & 0xffL) << 8) 
	            | (((long)data[offset]) & 0xffL);  	    	
    }

    /**
     * Read 2 bytes of data representing a signed 16 bit short value stored as little endian. 
     * @param data the byte array where the int16 has to be extracted
     * @param offset the offset in the byte array of the first byte of the int16 value
     * @return the int16 value as short value
     */
    public static short readShortLittleEndian(byte[] data, int offset){
	    return (short) ( ((data[offset+1] & 0xff) << 8) | (data[offset] & 0xff) );
    }
    
    /**
     * Read 2 bytes of data representing a signed 16 bit short value stored as big endian. 
     * @param data the byte array where the int16 has to be extracted
     * @param offset the offset in the byte array of the first byte of the int16 value
     * @return the int16 value as short value
     */
    public static short readShortBigEndian(byte[] data, int offset){
	    return (short) ( ((data[offset] & 0xff) << 8) | (data[offset+1] & 0xff) );
    }
    
    /**
     * Read 2 bytes of data representing a unsigned 16 bit short value stored as little endian. 
     * Because Java only supports signed values the return type is int instead of short.
     * @param data the byte array where the uint16 has to be extracted
     * @param offset the offset in the byte array of the first byte of the uint16 value
     * @return the int16 value as int value
     */
    public static int readUnsignedShortLittleEndian(byte[] data, int offset){
	    return  ( (( data[offset + 1] & 0xff) << 8) | (data[offset] & 0xff) );
    }

    /**
     * Read 2 bytes of data representing a unsigned 16 bit short value stored as big endian. 
     * Because Java only supports signed values the return type is int instead of short.
     * @param data the byte array where the uint16 has to be extracted
     * @param offset the offset in the byte array of the first byte of the uint16 value
     * @return the int16 value as int value
     */  
    public static int readUnsignedShortBigEndian(byte[] data, int offset){
	    return ( ((data[offset] & 0xff) << 8) | (data[offset+1] & 0xff) );
    }

}
