package de.uni_mannheim.aware.util;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Vector;

import javax.microedition.io.Connector;
import javax.microedition.io.HttpConnection;
import javax.microedition.location.Coordinates;

import me.regexp.RE;

import org.json.me.JSONArray;
import org.json.me.JSONException;
import org.json.me.JSONObject;

import de.uni_mannheim.aware.api.location.LocationData;
import de.uni_mannheim.aware.api.wireless.WirelessAccessPoint;
import de.uni_mannheim.aware.util.StringUtil;

/**
 * @see http://code.google.com/apis/gears/geolocation_network_protocol.html
 * @author Philipp Heckel
 *
 */
public class GeoUtil {
	/**
	 * The host of the web page that is requesting the location.
	 * This field is required by the protocol.
	 * 
	 * @see http://code.google.com/apis/gears/geolocation_network_protocol.html#request
	 */
	private static final String REQUEST_HOST = "uni-mannheim.de";
	
	/**
	 * An identifier which is set by the server and included
	 * in all subsequent requests to that server. Specified in the protocol.
	 * 
	 * @see http://code.google.com/apis/gears/geolocation_network_protocol.html#request
	 */
	private static String geolocAccessToken;
	
	private GeoUtil() {
	}

	public static LocationData test() throws IOException {
		Vector accessPoints = new Vector();
		
		accessPoints.addElement(new WirelessAccessPoint(new byte[] { 0x00, 0x21, 0x29, (byte) 0x83, 0x61, 0x1E}, "Lalala", -65));
		accessPoints.addElement(new WirelessAccessPoint(new byte[] { 0x00, 0x23, 0x08, 0x50, (byte) 0xD3, 0x21}, "EasyBox-Soner", -69));
		accessPoints.addElement(new WirelessAccessPoint(new byte[] { 0x00, 0x19, (byte) 0xCB, (byte) 0x86, 0x7E, 0x2E}, "PhiLeMiKa", -42));
		
		return getLocation(accessPoints);
	}
	
	/**
	 * Get location data using the Google Geolocation API, i.e. by sending available
	 * wireless access points to Google.
	 * 
	 * @param accessPoints Vector of {@link WirelessAccessPoint} objects.
	 * @return
	 * @throws IOException
	 */
	public static LocationData getLocation(Vector accessPoints) throws IOException {
		///// 1. Prepare request (create JSON) //////////////////////////////////////////////		
		String requestBody = "";
		
		// Create JSONArray for wifi towers
		JSONArray wifiTowers = new JSONArray();
		
		for (Enumeration e = accessPoints.elements(); e.hasMoreElements(); ) {
			WirelessAccessPoint wAP = (WirelessAccessPoint) e.nextElement();
			JSONObject accessPoint = new JSONObject();
			
			try {
				accessPoint.put("mac_address", StringUtil.getHexString(wAP.getBssid(),"-"));
				accessPoint.put("signal_strength", wAP.getSignalStrength());
			}
			catch (JSONException ex) {
				continue;
			}			
			
			wifiTowers.put(accessPoint);
		}
		
		// Create JSONObject request
		JSONObject request = new JSONObject();

		try {
			request.put("version", "1.1.0");
			request.put("host", REQUEST_HOST); // some host
			request.put("wifi_towers", wifiTowers);
			
			if (geolocAccessToken != null)
				request.put("access_token", geolocAccessToken);
		}
		catch (JSONException e) {
			// Nothing. Error cannot happen!
		}				
		
		requestBody = request.toString();

		System.out.println(requestBody);

		
		///// 2. Send request and receive answer //////////////////////////////////////////////		

		// Open connection
		HttpConnection http = 
			(HttpConnection) Connector.open("http://www.google.com/loc/json", Connector.READ_WRITE);
		
		http.setRequestMethod(HttpConnection.POST);
		http.setRequestProperty("Content-Type", "application/json");
		
		// Send request
		DataOutputStream dos = http.openDataOutputStream();  
		byte[] byteRequest = requestBody.getBytes(); 
		
		for(int i = 0; i<byteRequest.length; i++) 
		       dos.writeByte(byteRequest[i]);

		//dos.flush(); //Including this line may produce undesiredresults on certain devices
		/* If you call c.flush(), some devices will send the request in HTTP 1.1 "chunked" encoding by adding a "Transfer-encoding: chunked" header to the request and writing the size of each block of data in hexadecimal form as a text line before the data bytes themselves. This is repeated for any number of blocks you send, usually demarcated by calls to flush().
		If you have a HTTP 1.1-compliant Web server this should be transparent to your server side scripts, but if you do not, you will either want to avoid calling flush() (though some devices may still chunk for you automatically) or write some HTTP 1.1 chunk handling code in your server scripts. All you will need to do in the latter case is read the first text line of the response, extract the size from it (i.e. "C\r\n" where C means 12 bytes of chunk follow), and then read 12 bytes of response data. Then read another line to see if there is another chunk size, and repeat until stream is closed or your script has all the data it needs. See the HTTP 1.1 specification for more details.  */
				
		// Receive answer
		StringBuffer responseBody = new StringBuffer();		
		InputStream dis = http.openInputStream();
		
		int ch;  
		while ((ch = dis.read()) != -1)
			responseBody.append((char) ch);

		System.out.println(responseBody.toString());

		
		// Close connection
		dos.close();
		dis.close();
		http.close(); 

		
		///// 3. Evaluate/Parse response //////////////////////////////////////////////		

		try {
			JSONObject response = new JSONObject(responseBody.toString());
			
			// Parse
			if (response.has("access_token"))
				geolocAccessToken = response.getString("access_token");
			
			JSONObject location = response.getJSONObject("location");
			
			double latitude = location.getDouble("latitude");
			double longitude = location.getDouble("longitude");
						
			// Location object
			LocationData locationData = new LocationData(latitude, longitude);
			
			if (location.has("accuracy"))
				locationData.setHorizontalAccuracy(location.getDouble("accuracy"));

			if (location.has("altitude_accuracy"))
				locationData.setVerticalAccuracy(location.getDouble("altitude_accuracy"));

			return locationData;
		}
		catch (JSONException e) {
			throw new IOException();
		}	
	}
	
	/**
	 * Uses Google Maps to determine a time estimate to get from one location 
	 * to another. It returns the estimated minutes for the fastest route via car.
	 * If no route can be found or no connection is available, it throws an
	 * {@link IOException}.
	 * 
	 * <p><b>Note</b>: This method parses the output from Google Maps, i.e. 
	 * it might be necessary to adjust it if the output format changes.
	 * 
	 * <p>Implementation details: The method opens the following url and parses
	 * the JSON embedded in the HTML for a string that looks like <code>time:"..."</code>:
	 * http://maps.google.com/maps?f=q&source=s_q&hl=en&geocode=&q=from:LAT,LNG+to:LAT,LNG&ie=UTF8&t=h&z=12
	 * 
	 * @param source
	 * @param destination
	 * @return Estimated time (in minutes) for the fastest route via car
	 */
	public static int getTimeEstimate(LocationData source, LocationData destination) throws IOException {	
		String mapURL = 
			"http://maps.google.com/maps?f=q&source=s_q&hl=en&geocode=&ie=UTF8&t=h&z=12&q="+
			"from:"+source.getLatitude()+","+source.getLongitude()+"%20"+
			"to:"+destination.getLatitude()+","+destination.getLongitude();
		
		//System.out.println(mapURL);
		
		// Open connection, receive answer, close connection
		HttpConnection http = (HttpConnection) Connector.open(mapURL);
		String timeStr = StreamUtil.findNext("time:\"", "\"", http.openInputStream());		

		http.close(); 		
		
		// Parse answer
		RE re = new RE("((\\d+)\\s*hours?)?\\s*(\\d+)\\s*mins?");
		
		if (!re.match(timeStr))
			throw new IOException("Retrieved time string does not match (1): "+timeStr);
		
		int hours = 0;
		int mins = 0;
		
		String hourStr = re.getParen(2);
		String minStr = re.getParen(3);
		
		try {
			hours = (hourStr == null || "".equals(hourStr)) ? 0 : Integer.parseInt(hourStr);
			mins = (minStr == null || "".equals(minStr)) ? 0 : Integer.parseInt(minStr);
		}
		catch (NumberFormatException e) {
			// Cannot happen bc of "\d+"
			throw new IOException("Retrieved time string does not match (2): "+timeStr);
		}
		
		return hours*60+mins;
	}	
}
