package de.uni_mannheim.aware.util;

/* 
 * Ported from the Sun Microsystems FDLIBM C-library.
 * (Freely Distributable Library for Math)
 * ====================================================
 * Copyright (C) 2004 by Sun Microsystems, Inc. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice 
 * is preserved.
 * ====================================================
 */

/**
 * MathUtil for Java ME.  
 * This fills the gap in Java ME Math with a port of Sun's public FDLIBM C-library for IEEE-754.  
 * 
 * <p>Note: Extended by Team Aware by the methods {@link MathUtil#asin(double)},
 * {@link MathUtil#atan(double)}, {@link MathUtil#atan2(double, double)}; taken 
 * from: <a href="http://location-source.googlecode.com/svn/trunk/LocationSource/src/java/javax/microedition/location/Coordinates.java">Location Source, an implementation of the Java Location API (JSR-179)</a>
 * 
 * @author kmashint
 *
 * @see http://www.netlib.org/fdlibm/readme
 * 		For the Freely Distributable C-library conforming to IEEE-754 floating point math.
 * @see http://web.mit.edu/source/third/gcc/libjava/java/lang/
 *		For the GNU C variant of the same IEEE-754 routines.
 * @see http://www.dclausen.net/projects/microfloat/
 * 		Another take on the IEEE-754 routines.
 * @see http://real-java.sourceforge.net/Real.html
 * 		Yet another take on the IEEE-754 routines.
 * @see http://today.java.net/pub/a/today/2007/11/06/creating-java-me-math-pow-method.html
 * 		For other approximations.
 * @see http://martin.ankerl.com/2007/10/04/optimized-pow-approximation-for-java-and-c-c/
 * 		For fast but rough approximations.
 * @see http://martin.ankerl.com/2007/02/11/optimized-exponential-functions-for-java/
 * 		For more fast but rough approximations.
 */
public abstract class MathUtil {

	/* Common constants. */
	
	private static final double
		zero	=  0.0,
		one		=  1.0,
		huge	=  1.0e+300,
		two54   =  1.80143985094819840000e+16, /* 0x43500000, 0x00000000 */
		P1   =  1.66666666666666019037e-01, /* 0x3FC55555, 0x5555553E */
		P2   = -2.77777777770155933842e-03, /* 0xBF66C16C, 0x16BEBD93 */
		P3   =  6.61375632143793436117e-05, /* 0x3F11566A, 0xAF25DE2C */
		P4   = -1.65339022054652515390e-06, /* 0xBEBBBD41, 0xC5D26BF1 */
		P5   =  4.13813679705723846039e-08; /* 0x3E663769, 0x72BEA4D0 */
	
	
	private static final double
		log10 = 2.302585092994046D; /* Natural log(10.0D). */
	
	private static final long
		HI_MASK = 0xffffffff00000000L,
		LO_MASK = 0x00000000ffffffffL;
	
	private static final int
		HI_SHIFT = 32;
	
	
	/**
	 * Return Math.E to the exponent a.
	 * This in turn uses ieee7854_exp(double). 
	 */
	public static final double exp(double a) {
		return ieee754_exp(a);
	}
	
	/**
	 * Return the natural 
arithm, ln(a), as it relates to Math.E.
	 * This in turn uses ieee7854_log(double).
	 */
	public static final double log(double a) {
		return ieee754_log(a);
	}
	
	/**
	 * Return the common base-10 logarithm, log10(a).
	 * This in turn uses ieee7854_log(double)/ieee7854_log(10.0).
	 */
	public static final double log10(double a) {
		return ieee754_log(a)/log10;
	}


	/* __ieee754_exp(x)
	 * Returns the exponential of x.
	 *
	 * Method
	 *   1. Argument reduction:
	 *      Reduce x to an r so that |r| <= 0.5*ln2 ~ 0.34658.
	 *	Given x, find r and integer k such that
	 *
	 *               x = k*ln2 + r,  |r| <= 0.5*ln2.  
	 *
	 *      Here r will be represented as r = hi-lo for better 
	 *	accuracy.
	 *
	 *   2. Approximation of exp(r) by a special rational function on
	 *	the interval [0,0.34658]:
	 *	Write
	 *	    R(r**2) = r*(exp(r)+1)/(exp(r)-1) = 2 + r*r/6 - r**4/360 + ...
	 *      We use a special Remes algorithm on [0,0.34658] to generate 
	 * 	a polynomial of degree 5 to approximate R. The maximum error 
	 *	of this polynomial approximation is bounded by 2**-59. In
	 *	other words,
	 *	    R(z) ~ 2.0 + P1*z + P2*z**2 + P3*z**3 + P4*z**4 + P5*z**5
	 *  	(where z=r*r, and the values of P1 to P5 are listed below)
	 *	and
	 *	    |                  5          |     -59
	 *	    | 2.0+P1*z+...+P5*z   -  R(z) | <= 2 
	 *	    |                             |
	 *	The computation of exp(r) thus becomes
	 *                             2*r
	 *		exp(r) = 1 + -------
	 *		              R - r
	 *                                 r*R1(r)	
	 *		       = 1 + r + ----------- (for better accuracy)
	 *		                  2 - R1(r)
	 *	where
	 *			         2       4             10
	 *		R1(r) = r - (P1*r  + P2*r  + ... + P5*r   ).
	 *	
	 *   3. Scale back to obtain exp(x):
	 *	From step 1, we have
	 *	   exp(x) = 2^k * exp(r)
	 *
	 * Special cases:
	 *	exp(INF) is INF, exp(NaN) is NaN;
	 *	exp(-INF) is 0, and
	 *	for finite argument, only exp(0)=1 is exact.
	 *
	 * Accuracy:
	 *	according to an error analysis, the error is always less than
	 *	1 ulp (unit in the last place).
	 *
	 * Misc. info.
	 *	For IEEE double 
	 *	    if x >  7.09782712893383973096e+02 then exp(x) overflow
	 *	    if x < -7.45133219101941108420e+02 then exp(x) underflow
	 *
	 * Constants:
	 * The hexadecimal values are the intended ones for the following 
	 * constants. The decimal values may be used, provided that the 
	 * compiler will convert from decimal to binary accurately enough
	 * to produce the hexadecimal values shown.
	 */

	private static final double
		twom1000= 9.33263618503218878990e-302,     /* 2**-1000=0x01700000,0*/
		o_threshold=  7.09782712893383973096e+02,  /* 0x40862E42, 0xFEFA39EF */
		u_threshold= -7.45133219101941108420e+02,  /* 0xc0874910, 0xD52D3051 */
		invln2 =  1.44269504088896338700e+00; /* 0x3ff71547, 0x652b82fe */

	private static final double[]
	    halF	= new double[] { 0.5, -0.5 },
	    ln2HI   = new double[] { 6.93147180369123816490e-01,  /* 0x3fe62e42, 0xfee00000 */
			     -6.93147180369123816490e-01 }, /* 0xbfe62e42, 0xfee00000 */
		ln2LO   = new double[] { 1.90821492927058770002e-10,  /* 0x3dea39ef, 0x35793c76 */
		     -1.90821492927058770002e-10 }; /* 0xbdea39ef, 0x35793c76 */

	private static final double ieee754_exp(double x)
	{
		double y,c,t;
		double hi = 0, lo = 0;
		int k = 0 ;
		int xsb, hx, lx;
		long yl;
		long xl = Double.doubleToLongBits(x);

		hx  = (int)((long)xl >>> HI_SHIFT);	/* high word of x */
		xsb = (hx>>31)&1;		/* sign bit of x */
		hx &= 0x7fffffff;		/* high word of |x| */

	    /* filter out non-finite argument */
		if(hx >= 0x40862E42) {			/* if |x|>=709.78... */
	        if(hx>=0x7ff00000) {
	    	lx	= (int)((long)xl & LO_MASK);	/* low word of x */
			if(((hx&0xfffff)|lx)!=0) 
			     return x+x; 		/* NaN */
			else return (xsb==0)? x:0.0;	/* exp(+-inf)={inf,0} */
		    }
		    if(x > o_threshold) return huge*huge; /* overflow */
		    if(x < u_threshold) return twom1000*twom1000; /* underflow */
		}

	    /* argument reduction */
		if(hx > 0x3fd62e42) {		/* if  |x| > 0.5 ln2 */ 
		    if(hx < 0x3FF0A2B2) {	/* and |x| < 1.5 ln2 */
			hi = x-ln2HI[xsb]; lo=ln2LO[xsb]; k = 1-xsb-xsb;
		    } else {
			k  = (int)(invln2*x+halF[xsb]);
			t  = k;
			hi = x - t*ln2HI[0];	/* t*ln2HI is exact here */
			lo = t*ln2LO[0];
		    }
		    x  = hi - lo;
		} 
		else if(hx < 0x3e300000)  {	/* when |x|<2**-28 */
		    if(huge+x>one) return one+x;/* trigger inexact */
		}
		//else k = 0; // handled at declaration

		/* x is now in primary range */
		t  = x*x;
		c  = x - t*(P1+t*(P2+t*(P3+t*(P4+t*P5))));
		if(k==0) 	return one-((x*c)/(c-2.0)-x); 
		else 		y = one-((lo-(x*c)/(2.0-c))-hi);
		yl	 = Double.doubleToLongBits(y);
		if(k >= -1021) {
		    yl += ((long)k<<(20+HI_SHIFT));	/* add k to y's exponent */
		    return Double.longBitsToDouble(yl);
		} else {
		    yl += ((long)(k+1000)<<(20+HI_SHIFT));/* add k to y's exponent */
		    return Double.longBitsToDouble(yl)*twom1000;
		}
	}

	/* __ieee754_log(x)
	 * Return the logrithm of x
	 *
	 * Method :                  
	 *   1. Argument Reduction: find k and f such that 
	 *			x = 2^k * (1+f), 
	 *	   where  sqrt(2)/2 < 1+f < sqrt(2) .
	 *
	 *   2. Approximation of log(1+f).
	 *	Let s = f/(2+f) ; based on log(1+f) = log(1+s) - log(1-s)
	 *		 = 2s + 2/3 s**3 + 2/5 s**5 + .....,
	 *	     	 = 2s + s*R
	 *      We use a special Reme algorithm on [0,0.1716] to generate 
	 * 	a polynomial of degree 14 to approximate R The maximum error 
	 *	of this polynomial approximation is bounded by 2**-58.45. In
	 *	other words,
	 *		        2      4      6      8      10      12      14
	 *	    R(z) ~ Lg1*s +Lg2*s +Lg3*s +Lg4*s +Lg5*s  +Lg6*s  +Lg7*s
	 *  	(the values of Lg1 to Lg7 are listed in the program)
	 *	and
	 *	    |      2          14          |     -58.45
	 *	    | Lg1*s +...+Lg7*s    -  R(z) | <= 2 
	 *	    |                             |
	 *	Note that 2s = f - s*f = f - hfsq + s*hfsq, where hfsq = f*f/2.
	 *	In order to guarantee error in log below 1ulp, we compute log
	 *	by
	 *		log(1+f) = f - s*(f - R)	(if f is not too large)
	 *		log(1+f) = f - (hfsq - s*(hfsq+R)).	(better accuracy)
	 *	
	 *	3. Finally,  log(x) = k*ln2 + log(1+f).  
	 *			    = k*ln2_hi+(f-(hfsq-(s*(hfsq+R)+k*ln2_lo)))
	 *	   Here ln2 is split into two floating point number: 
	 *			ln2_hi + ln2_lo,
	 *	   where n*ln2_hi is always exact for |n| < 2000.
	 *
	 * Special cases:
	 *	log(x) is NaN with signal if x < 0 (including -INF) ; 
	 *	log(+INF) is +INF; log(0) is -INF with signal;
	 *	log(NaN) is that NaN with no signal.
	 *
	 * Accuracy:
	 *	according to an error analysis, the error is always less than
	 *	1 ulp (unit in the last place).
	 *
	 * Constants:
	 * The hexadecimal values are the intended ones for the following 
	 * constants. The decimal values may be used, provided that the 
	 * compiler will convert from decimal to binary accurately enough 
	 * to produce the hexadecimal values shown.
	 */

	private static final double
		ln2_hi  =  6.93147180369123816490e-01,	/* 3fe62e42 fee00000 */
		ln2_lo  =  1.90821492927058770002e-10,	/* 3dea39ef 35793c76 */
		Lg1 = 6.666666666666735130e-01,  /* 3FE55555 55555593 */
		Lg2 = 3.999999999940941908e-01,  /* 3FD99999 9997FA04 */
		Lg3 = 2.857142874366239149e-01,  /* 3FD24924 94229359 */
		Lg4 = 2.222219843214978396e-01,  /* 3FCC71C5 1D8E78AF */
		Lg5 = 1.818357216161805012e-01,  /* 3FC74664 96CB03DE */
		Lg6 = 1.531383769920937332e-01,  /* 3FC39A09 D078C69F */
		Lg7 = 1.479819860511658591e-01;  /* 3FC2F112 DF3E5244 */

	private static final double ieee754_log(double x)
	{
		double hfsq,f,s,z,R,w,t1,t2,dk;
		int k,hx,lx,i,j;
		long xl = Double.doubleToLongBits(x);

		hx = (int)(xl >> HI_SHIFT);		/* high word of x */
		lx = (int)(xl & LO_MASK);		/* low  word of x */

		k=0;
		if (hx < 0x00100000) {			/* x < 2**-1022  */
		    if (((hx&0x7fffffff)|lx)==0) 
			return -two54/zero;		/* log(+-0)=-inf */
		    if (hx<0) return (x-x)/zero;	/* log(-#) = NaN */
		    k -= 54; x *= two54; /* subnormal number, scale up x */
		    hx = (int)(Double.doubleToLongBits(x) >>> HI_SHIFT);		/* high word of x */
		} 
		if (hx >= 0x7ff00000) return x+x;
		k += (hx>>20)-1023;
		hx &= 0x000fffff;
		i = (hx+0x95f64)&0x100000;
		//__HI(x) = hx|(i^0x3ff00000);	/* normalize x or x/2 */
		x = Double.longBitsToDouble(((long)(hx|(i^0x3ff00000)) << HI_SHIFT) | (Double.doubleToLongBits(x) & LO_MASK));
		k += (i>>20);
		f = x-1.0;
		if((0x000fffff&(2+hx))<3) {	/* |f| < 2**-20 */
		    if(f==zero) if(k==0) return zero;  else {dk=(double)k;
					 return dk*ln2_hi+dk*ln2_lo;}
		    R = f*f*(0.5-0.33333333333333333*f);
		    if(k==0) return f-R; else {dk=(double)k;
		    	     return dk*ln2_hi-((R-dk*ln2_lo)-f);}
		}
	 	s = f/(2.0+f); 
		dk = (double)k;
		z = s*s;
		i = hx-0x6147a;
		w = z*z;
		j = 0x6b851-hx;
		t1= w*(Lg2+w*(Lg4+w*Lg6)); 
		t2= z*(Lg1+w*(Lg3+w*(Lg5+w*Lg7))); 
		i |= j;
		R = t2+t1;
		if(i>0) {
		    hfsq=0.5*f*f;
		    if(k==0) return f-(hfsq-s*(hfsq+R)); else
			     return dk*ln2_hi-((hfsq-(s*(hfsq+R)+dk*ln2_lo))-f);
		} else {
		    if(k==0) return f-s*(f-R); else
			     return dk*ln2_hi-((s*(f-R)-dk*ln2_lo)-f);
		}
	}

	
	
	// EXTENDED //////////////////////////////////////////////////////////////////////////////////////////
	
	/// FROM 
	//////   http://location-source.googlecode.com/svn/trunk/LocationSource/src/java/javax/microedition/location/Coordinates.java
	//////   http://www.google.com/codesearch/p?hl=en#CcCT9_yRf_w/trunk/LocationSource/src/java/javax/microedition/location/Coordinates.java&q=public%20class%20javax.microedition.location.Coordinates%20lang:java&sa=N&cd=5&ct=rc
	
	
	
	
    /**
     * Constant for PI divided by 2.
     */
	public static final double PIover2 = Math.PI / 2;

    /**
     * Constant for PI divided by 4.
     */
    public static final double PIover4 = Math.PI / 4;

    /**
     * Constant for PI divided by 6.
     */
    public static final double PIover6 = Math.PI / 6;

    /**
     * Constant for PI divided by 12.
     */
    public static final double PIover12 = Math.PI / 12;

    /**
     * Constant used in the <code>atan</code> calculation.
     */
    public static final double ATAN_CONSTANT = 1.732050807569;


	
	
    /**
     * Returns the arc tangent of an angle, in the range of <code>-Math.PI/2</code>
     * through <code>Math.PI/2</code>.  Special cases:
     * <ul>
     *  <li>If the argument is <code>NaN</code>, then the result is <code>NaN</code>.
     *  <li>If the argument is zero, then the result is a zero with the same
     *      sign as the argument.
     * </ul>
     * <p>
     * A result must be within 1 ulp of the correctly rounded result.  Results
     * must be semi-monotonic.
     *
     * @param a - the value whose arc tangent is to be returned.
     * @return the arc tangent of the argument.
     */
    public static double atan (double a)
    {
            // Special cases.
            if ( Double.isNaN(a) )
            {
                    return Double.NaN;
            }

            if ( a == 0.0 )
            {
                    return a;
            }

            // Compute the arc tangent.
            boolean negative = false;
            boolean greaterThanOne = false;
            int i = 0;

            if ( a < 0.0 )
            {
                    a = -a;
                    negative = true;
            }

            if ( a > 1.0 )
            {
                    a = 1.0 / a;
                    greaterThanOne = true;
            }

            double t;

            for ( ; a > PIover12; a *= t )
            {
                    i++;
                    t = a + ATAN_CONSTANT;
                    t = 1.0 / t;
                    a *= ATAN_CONSTANT;
                    a--;
            }

            double aSquared = a * a;

            double arcTangent = aSquared + 1.4087812;
            arcTangent = 0.55913709 / arcTangent;
            arcTangent += 0.60310578999999997;
            arcTangent -= 0.051604539999999997 * aSquared;
            arcTangent *= a;

            for ( ; i > 0; i-- )
            {
                    arcTangent += PIover6;
            }

            if ( greaterThanOne )
            {
                    arcTangent = PIover2 - arcTangent;
            }

            if ( negative )
            {
                    arcTangent = -arcTangent;
            }

            return arcTangent;
    }

	
    /**
     * Returns the arc sine of an angle, in the range of <code>-Math.PI/2</code> through
     * <code>Math.PI/2</code>.  Special cases:
     * <ul>
     *  <li>If the argument is <code>NaN</code> or its absolute value is greater than 1,
     *      then the result is <code>NaN</code>.
     *  <li>If the argument is zero, then the result is a zero with the same sign
     *      as the argument.
     * </ul>
     *
     * @param a - the value whose arc sine is to be returned.
     * @return the arc sine of the argument.
     */
    public static double asin (double a)
    {
            // Special cases.
            if ( Double.isNaN(a) || Math.abs(a) > 1.0 )
            {
                    return Double.NaN;
            }

            if ( a == 0.0 )
            {
                    return a;
            }

            // Calculate the arc sine.
            double aSquared = a * a;
            double arcSine = atan2( a, Math.sqrt(1 - aSquared) );
            return arcSine;
    }

    /**
     * Converts rectangular coordinates (x, y) to polar (r, <i>theta</i>).  This method
     * computes the phase <i>theta</i> by computing an arc tangent of y/x in the range
     * of <i>-pi</i> to <i>pi</i>.  Special cases:
     * <ul>
     *  <li>If either argument is <code>NaN</code>, then the result is <code>NaN</code>.
     *  <li>If the first argument is positive zero and the second argument is
     *      positive, or the first argument is positive and finite and the second
     *      argument is positive infinity, then the result is positive zero.
     *  <li>If the first argument is negative zero and the second argument is
     *      positive, or the first argument is negative and finite and the second
     *      argument is positive infinity, then the result is negative zero.
     *  <li>If the first argument is positive zero and the second argument is
     *      negative, or the first argument is positive and finite and the second
     *      argument is negative infinity, then the result is the <code>double</code> value
     *      closest to <i>pi</i>.
     *  <li>If the first argument is negative zero and the second argument is
     *      negative, or the first argument is negative and finite and the second
     *      argument is negative infinity, then the result is the <code>double</code> value
     *      closest to <i>-pi</i>.
     *  <li>If the first argument is positive and the second argument is positive
     *      zero or negative zero, or the first argument is positive infinity and
     *      the second argument is finite, then the result is the <code>double</code> value
     *      closest to <i>pi</i>/2.
     *  <li>If the first argument is negative and the second argument is positive
     *      zero or negative zero, or the first argument is negative infinity and
     *      the second argument is finite, then the result is the <code>double</code> value
     *      closest to <i>-pi</i>/2.
     *  <li>If both arguments are positive infinity, then the result is the double
     *      value closest to <i>pi</i>/4.
     *  <li>If the first argument is positive infinity and the second argument is
     *      negative infinity, then the result is the double value closest to 3*<i>pi</i>/4.
     *  <li>If the first argument is negative infinity and the second argument is
     *      positive infinity, then the result is the double value closest to -<i>pi</i>/4.
     *  <li>If both arguments are negative infinity, then the result is the double
     *      value closest to -3*<i>pi</i>/4.
     * </ul>
     * <p>
     * A result must be within 2 ulps of the correctly rounded result.  Results
     * must be semi-monotonic.
     *
     * @param y - the ordinate coordinate
     * @param x - the abscissa coordinate
     * @return the <i>theta</i> component of the point (r, <i>theta</i>) in polar
     *   coordinates that corresponds to the point (x, y) in Cartesian coordinates.
     */
    public static double atan2 (double y, double x)
    {
            // Special cases.
            if ( Double.isNaN(y) || Double.isNaN(x) )
            {
                    return Double.NaN;
            }
            else if ( Double.isInfinite(y) )
            {
                    if ( y > 0.0 ) // Positive infinity
                    {
                            if ( Double.isInfinite(x) )
                            {
                                    if ( x > 0.0 )
                                    {
                                            return PIover4;
                                    }
                                    else
                                    {
                                            return 3.0 * PIover4;
                                    }
                            }
                            else if ( x != 0.0 )
                            {
                                    return PIover2;
                            }
                    }
                    else  // Negative infinity
                    {
                            if ( Double.isInfinite(x) )
                            {
                                    if ( x > 0.0 )
                                    {
                                            return -PIover4;
                                    }
                                    else
                                    {
                                            return -3.0 * PIover4;
                                    }
                            }
                            else if ( x != 0.0 )
                            {
                                    return -PIover2;
                            }
                    }
            }
            else if ( y == 0.0 )
            {
                    if ( x > 0.0 )
                    {
                            return y;
                    }
                    else if ( x < 0.0 )
                    {
                            return Math.PI;
                    }
            }
            else if ( Double.isInfinite(x) )
            {
                    if ( x > 0.0 )  // Positive infinity
                    {
                            if ( y > 0.0 )
                            {
                                    return 0.0;
                            }
                            else if ( y < 0.0 )
                            {
                                    return -0.0;
                            }
                    }
                    else  // Negative infinity
                    {
                            if ( y > 0.0 )
                            {
                                    return Math.PI;
                            }
                            else if ( y < 0.0 )
                            {
                                    return -Math.PI;
                            }
                    }
            }
            else if ( x == 0.0 )
            {
                    if ( y > 0.0 )
                    {
                            return PIover2;
                    }
                    else if ( y < 0.0 )
                    {
                            return -PIover2;
                    }
            }


            // Implementation a simple version ported from a PASCAL implementation:
            //   http://everything2.com/index.pl?node_id=1008481

            double arcTangent;

            // Use arctan() avoiding division by zero.
            if ( Math.abs(x) > Math.abs(y) )
            {
                    arcTangent = atan(y / x);
            }
            else
            {
                    arcTangent = atan(x / y); // -PI/4 <= a <= PI/4

                    if ( arcTangent < 0 )
                    {
                            arcTangent = -PIover2 - arcTangent; // a is negative, so we're adding
                    }
                    else
                    {
                            arcTangent = PIover2 - arcTangent;
                    }
            }

            // Adjust result to be from [-PI, PI]
            if ( x < 0 )
            {
                    if ( y < 0 )
                    {
                            arcTangent = arcTangent - Math.PI;
                    }
                    else
                    {
                            arcTangent = arcTangent + Math.PI;
                    }
            }

            return arcTangent;
    }


	
	
	
}
