package de.uni_mannheim.aware.util;

import java.io.ByteArrayInputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;

/**
 * StreamUtil comprises several I/O stream oriented functions, especially
 * {@link StreamUtil#findNext(String, String, InputStream) findNext()}, a method
 * to parse information from an {@link InputStream}. 
 * 
 * @author Philipp Heckel
 */
public class StreamUtil {
	/**
	 * Skips through an input stream until the given search string is found. If the 
	 * string is found, the method returns <em>true</em>. The current position is the
	 * first character after the search string. If the string is not found, it seeks to
	 * the end of the input stream and throws an EOFException.
	 * 
	 * <p><b>Example</b>:<br />
	 *   The search string <b>&lt;b&gt;</b> on the input data <b>Hi &lt;b&gt;Bob&lt;/b&gt;</b>
	 *   returns true and points to <b>B</b> (assuming that the current stream position is 0.
	 * 
	 * @param searchStr String to search for in the input stream.
	 * @param is Input stream to skip through
	 * @return TRUE if the needle was found, FALSE if the EOF has been reached
	 * @throws EOFException If the method reaches the end of the input stream
	 *         before the search string was found. 
	 * @throws IOException If the socket breaks, or any other I/O exception occurs
	 */
	public static boolean seekAfterMatch(String searchStr, InputStream is) throws EOFException, IOException { 
		char[] matchStr = searchStr.toCharArray();
		int matchPos = 0;
		
		int c;
		while ((c = is.read()) != -1 /* not EOF */) {
			
			// increase match position if next char matches
			if ((char) c == matchStr[matchPos]) {
				matchPos++;
				//System.out.println("+:"+(char) c);
				
				// break if last
				if (matchPos == matchStr.length)
					break;
			}
			
			// reset otherwise
			else {
				matchPos = 0;
				//System.out.println("-:"+(char) c);
			}
				
		}
		
		// No match (EOF)
		if (matchPos != matchStr.length)
			throw new EOFException("Search string '"+searchStr+"' not found, EOF reached.");
		
		// Match
		return true;		
	}
	
	/**
	 * Reads from the given input stream until the search string is found. If the string is found,
	 * the function returns the string read up to the first character of the search string. If it is
	 * not found, it seeks to the end of the input stream and throws a EOFException.
	 * 
	 * <p><b>Example</b>:<br />
	 *   The search string <b>&lt;/b&gt;</b> on the input data <b>Hi &lt;b&gt;Bob&lt;/b&gt;</b>
	 *   returns <b>Hi &lt;b&gt;Bob</b> if the current position of the stream is 0. It the current
	 *   position is 6 (after the &lt;b&gt;), it returns <b>Bob</b>.
	 *   
	 * @param searchStr String to search for in the input stream.
	 * @param inputStream Input stream to skip through
	 * @return String between the current position of the stream and the first character of the search string
	 *         in the input stream
	 * @throws EOFException If the method reaches the end of the input stream
	 *         before the search string was found. 
	 * @throws IOException If the socket breaks, or any other I/O exception occurs
	 */
	public static String readUntil(String searchStr, InputStream inputStream) throws EOFException, IOException {
		StringBuffer response = new StringBuffer();

		char[] matchStr = searchStr.toCharArray();
		int matchPos = 0;
				
		int c;
		while ((c = inputStream.read()) != -1 /* not EOF */) {			
			// increase match position if next char matches
			if ((char) c == matchStr[matchPos]) {
				matchPos++;
				//System.out.println("Y:"+(char) c);		
				
				// break if last
				if (matchPos == matchStr.length)
					break;
			}
			
			// otherwise
			else {
				// add parts of match that haven't been added
				// yet (could have been a match)
				if (matchPos > 0)
					response.append(searchStr.substring(0,matchPos));

				// Add current character
				response.append((char) c);
				//System.out.println("N:"+(char) c);			
				
				// reset match position counter
				matchPos = 0;						
			}
		}		

		// No match (EOF): 'match'-variable is to EOF
		if (matchPos != matchStr.length)
			throw new EOFException("Delimiter '"+searchStr+"' not found, EOF reached.");
		
		// Match: 'match'-variable is to 'delimiter'
		return response.toString();			
	}
	
	/**
	 * Finds and returns the first occurrence of string between two delimiters
	 * in the input stream. The method seeks to the first occurrence of the
	 * start delimiter and reads until the end delimiter is found. The method
	 * uses {@link StreamUtil#seekAfterMatch(String, InputStream) seekAfterMatch()}
	 * and {@link StreamUtil#readUntil(String, InputStream) readUntil()}.
	 * 
	 * <p><b>Example</b>:<br />
	 * <pre>
	 *   InputStream is = 
	 *      new ByteArrayInputStream("Hi &lt;b&gt;Alice&lt;/b&gt; and &lt;b&gt;Bob&lt;/b&gt;!".getBytes());
	 * 		
	 *   StreamUtil.findNext("&lt;b&gt;", "&lt;/b&gt;", is); // Alice
	 *   StreamUtil.findNext("&lt;b&gt;", "&lt;/b&gt;", is); // Bob
	 *   StreamUtil.findNext("&lt;b&gt;", "&lt;/b&gt;", is); // EOFException
	 * </pre>
	 * 
	 * @param startDelimiter Left delimiter of the string to parse from the input stream
	 * @param endDelimiter Right delimiter of the string to parse from the input stream
	 * @param inputStream Input stream to skip through
	 * @return String between the first occurrence of the left delimiter
	 *         and the next right delimiter in the input stream
	 * @throws EOFException If the method reaches the end of the input stream
	 *         before the search string was found. 
	 * @throws IOException If the socket breaks, or any other I/O exception occurs
	 */
	public static String findNext(String startDelimiter, String endDelimiter, InputStream inputStream) 
		throws EOFException, IOException {
		
		seekAfterMatch(startDelimiter, inputStream);
		return readUntil(endDelimiter, inputStream);
	}
	
	public static void main(String a[]) {
		InputStream is = new ByteArrayInputStream("Hi <b>Alice</b> and <b>Bob</b>!".getBytes());
		
		try {
			System.out.println(StreamUtil.findNext("<b>", "</b>", is));
			System.out.println(StreamUtil.findNext("<b>", "</b>", is));
			System.out.println(StreamUtil.findNext("!", "</tag>", is));
		}
		catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
	
}
